"""
Multipatch Spline Configuration
"""

import numpy as _np

from splinepy import settings as _settings
from splinepy._base import SplinepyBase as _SplinepyBase
from splinepy.helpme import visualize as _visualize
from splinepy.helpme.extract import Extractor as _Extractor
from splinepy.spline import _default_if_none, _get_helper
from splinepy.splinepy_core import PyMultipatch as _PyMultipatch
from splinepy.splinepy_core import (
    boundaries_from_continuity as _boundaries_from_continuity,
)
from splinepy.utils.data import MultipatchData as _MultipatchData


class Multipatch(_SplinepyBase, _PyMultipatch):
    """
    System of patches to store information such as boundaries and
    interfaces
    """

    __slots__ = (
        "_extractor",
        "_show_options",
        "_spline_data",
    )

    __show_option__ = _visualize.SplineShowOption

    def __init__(self, splines=None, interfaces=None, *, spline=None):
        """
        Multipatch

        Parameters
        ----------
        splines : list-like
          List of splines to store as multipatch
        interfaces : array-like
          Defines the connectivity in between patches
        spline : PyMultipatch
          keyword only argument, implemented to support to_derived() interface.
          calls move constructor.

        Returns
        -------
        None
        """
        # Init values
        if spline is not None:
            if not isinstance(spline, _PyMultipatch):
                raise TypeError("spline must be PyMultipatch.")
            super().__init__(spline)
        elif splines is not None:
            super().__init__(splines, _settings.NTHREADS, False)
        else:
            super().__init__()

        self._logd("Initialized Multipatch object")

        if interfaces is not None:
            self.interfaces = interfaces

    @property
    def patches(self):
        """
        List of splines in splinepy format

        Returns
        -------
        patches : list
         list of splines that are stored in the multipatch system


        """
        return super().patches

    @patches.setter
    def patches(self, list_of_splines):
        _PyMultipatch.patches.fset(self, list_of_splines)

    @property
    def interfaces(self):
        """
        Determine the interfaces of a given multipatch system

        Returns its interfaces in the form as an array of size
        n_patches x n_sides_per_patch

        The interfaces are stored as indices. If the index is negative, the ID
        is referring to a boundary (e.g. for boundary conditions). If the
        number is positive, the ID refers to the global face-id, which is
        composed of the element ID and the local face ID :
        global_face = local_face + number_of_faces_per_element * element_id

        Example:

        .. code-block ::

          O -- 3 -- O O -- 3 -- O
          |         | |         |
          0    0    1 0    1    1
          |         | |         |
          O -- 2 -- O O -- 2 -- O

        would be stored as:

        .. code-block ::

          [
            [-1, 4, -1, -1],
            [1, -1, -1, -1]
          ]

        Returns
        -------
        interfaces : np.ndarray
          inter-patch connectivitiy and boundaries
        """
        # empty list as input will compute interfaces based on input
        return super().get_interfaces(False)

    @interfaces.setter
    def interfaces(self, con):
        """super() checks validity of input"""
        # Assignment
        super().set_interfaces(con)

    def interface_orientations(self, tolerance=None, nthreads=None):
        """
        Determine orientation between adjacent splines which is required, e.g.,
        for gismo export and orientation check

        Parameters
        ----------
        tolerance : double
          normed distance between two neighboring points to be considered equal
        nthreads : int
          Number of threads to be used for extraction, defaults to
          settings.NTHREADS

        Returns
        -------
        interface_orientations : np.array (np.int64)
          description of the orientations between adjacent splines in a
          multipatch object
        """
        return super().interface_orientations(
            tolerance=_default_if_none(tolerance, _settings.TOLERANCE),
            nthreads=_default_if_none(nthreads, _settings.NTHREADS),
        )

    @property
    def boundaries(self):
        """
        Boundaries are stored within the interfaces array as negative entries

        Negative entries mean, that there is a boundary, the absolute value
        holds the boundary ID

        Returns
        -------
        boundaries : list<np.ndarray>
          list of 2D arrays, representing the patch and face ID
        """
        # Get minimum boundary id - boundary ids are stored as negative values
        max_BID = self.interfaces.min()

        boundary_list = []
        for i_bid in range(-1, max_BID - 1, -1):
            self._logd(f"Extracting boundary with ID {abs(i_bid)}")
            boundary_list.append(_np.where(self.interfaces == i_bid))
            self._logd(
                f"Found {boundary_list[-1][1].size} boundary "
                f"elements on boundary {abs(i_bid)}"
            )

        return boundary_list

    def boundary_multipatch(self, bid=-1, nthreads=None):
        """Extract all boundary patches of a given Multipatch system as splines

        Parameters
        ----------
        bid : int
          boundary id to extract, (default all if bid<0)
        nthreads : int
          Number of threads to be used for extraction, defaults to
          settings.NTHREADS

        Returns
        -------
        boundary_multipatch : Multipatch
          Embedded splines representing boundaries of system
        """
        if nthreads is None:
            nthreads = _settings.NTHREADS

        # apply nthreads
        previous_nthreads = self.n_default_threads
        self.n_default_threads = nthreads

        b_patches = super().boundary_multipatch(bid)

        self.n_default_threads = previous_nthreads

        return b_patches

    def set_boundary(
        self,
        spline_ids,
        boundary_faces,
        boundary_id=None,
    ):
        """
        Adds a boundary to a specific set of spline and spline-

        Parameters
        ----------
        spline_ids : np.ndarray
          array of spline-ids that are assigned a boundary
        boundary_faces : np.ndarray
          array of corresponding faces to be assigned a boundary
        boundary_id : int
          boundary_id to be assigned. If not chosen set to new lowest value

        Returns
        -------
        None
        """
        # Get minimum boundary id - boundary ids are stored as negative values
        max_BID = self.interfaces.min()

        if boundary_id is None:
            new_BID = max_BID - 1
            self._logd(f"Creating new boundary with ID {abs(new_BID)}")
        else:
            # Make sure its negative
            new_BID = -abs(int(boundary_id))
            if new_BID < max_BID:
                self._logd(f"Creating new boundary with ID {abs(new_BID)}")
            else:
                self._logd(
                    "Adding new boundary elements to existing "
                    f"boundary new boundary with ID {abs(new_BID)}"
                )

        try:
            old_indices = self.interfaces[spline_ids, boundary_faces]

        except BaseException:
            raise ValueError(
                "spline_ids and boundary_faces need to be one-dimensional, "
                "array-like objects to access 2D interface numpy array"
            )

        # raise if any old indices include non-negative entries
        # as it indicates that it's not a boundary sub-patch
        if (old_indices > -1).any():
            raise ValueError(
                "One or more of the assigned boundary elements are not"
                " on the patch surface, please check topology"
            )

        # set boundary id
        self.interfaces[spline_ids, boundary_faces] = new_BID

    @property
    def para_dim(self):
        """
        Parametric dimension of the splineset

        Returns
        -------
        para_dim : int
          Parametric dimensionality of the multipatch system
        """
        return super().para_dim

    @property
    def dim(self):
        """
        Physical dimension of the splineset
        Returns
        -------
        dim : int
          Physical dimensionality of the multipatch system
        """
        return super().dim

    @property
    def sub_patch_centers(self):
        """
        Evaluated centers of the individual patches, used to determine
        connectivity and identify boundaries based on position

        Returns
        -------
        spline_centers : np.ndarray
          coordinates of the patch-boundary centers
        """
        return super().sub_patch_centers()

    def determine_interfaces(self, tolerance=None, recompute=False):
        """
        Retrieve interfaces info

        Stores new information as interfaces

        Parameters
        ----------
        tolerance : double
          normed distance between two neighboring points to be considered equal
        recompute : bool (False)
          Enforce recomputation of interfaces

        Returns
        -------
        interfaces : array-like (n_patch x n_boundaries)
        """
        if tolerance is None:
            tolerance = _settings.TOLERANCE

        # save previous default tol
        previous_tol = self.tolerance

        # set given tolerance and compute interfaces
        self.tolerance = tolerance

        # Enforces computation only if recompute or non-available
        interfaces = super().get_interfaces(recompute)

        self._logd("Successfully provided new interfaces using uff algorithm")

        # set default tol back
        self.tolerance = previous_tol

        return interfaces

    def boundary_from_function(
        self,
        function,
        mask=None,
        boundary_id=None,
    ):
        """
        Uses all faces, that were identified as boundaries in the
        interfaces array and checks if they fit a boundary function

        Parameters
        ----------
        function : Callable
          Function called on every boundary center point to check if it is on
          the boundary, returns bool-type
        mask : array-like
          If assigned, takes only boundaries with matching ids
        boundary_id : int
          boundary_id to be assigned. If not chosen set to new lowest value

        Returns
        -------
        None
        """
        # Get minimum boundary id - boundary ids are stored as negative values
        max_BID = self.interfaces.min()

        if boundary_id is None:
            new_BID = max_BID - 1
            self._logd(f"Creating new boundary with ID {abs(new_BID)}")
        else:
            # Make sure its negative
            new_BID = -abs(int(boundary_id))
            if new_BID < max_BID:
                self._logd(f"Creating new boundary with ID {abs(new_BID)}")
            else:
                self._logd(
                    "Adding new boundary elements to existing "
                    f"boundary new boundary with ID {abs(new_BID)}"
                )

        # retrieve all boundary elements
        if mask is None:
            boundary_ids = self.interfaces < 0
        else:
            boundary_ids = _np.isin(-self.interfaces, _np.abs(mask))

        # Check if there is a boundary
        if not boundary_ids.any():
            self._logd(
                "No boundary elements could be identified that match "
                "requirements"
            )
            return None

        # Check all face centers
        relevant_boundary_centers = self.sub_patch_centers[
            boundary_ids.flatten(), :
        ]

        # Cols and Rows
        row_ids, col_ids = _np.where(boundary_ids)

        # Function is applied to all face centers
        try:
            new_boundary_bools = function(relevant_boundary_centers)
        except BaseException:
            raise ValueError(
                "Function is not applicable to array. Function layout must"
                " f(array<n_face_points,dim>) -> bools<n_face_points>"
            )

        # Assign new boundary ID to interfaces array
        self.interfaces[
            row_ids[new_boundary_bools], col_ids[new_boundary_bools]
        ] = new_BID

    @property
    def spline_data(self):
        """
        Spline data helper for splines. @todo (does not do anything at the
        moment)

        Parameters
        ----------
        None

        Returns
        -------
        spline_data: MultipatchData
        """
        return _get_helper(self, "_spline_data", _MultipatchData)

    @property
    def show_options(self):
        """
        Show option manager for Multipatches.

        Parameters
        ----------
        None

        Returns
        -------
        show_options: SplineShowOption
        """
        return _get_helper(self, "_show_options", self.__show_option__)

    def showable(self, **kwargs):
        """Equivalent to

        .. code-block:: python

            splinepy.helpme.visualize.show(
                spline, return_showable=True, **kwargs
            )

        Parameters
        ----------
        kwargs: kwargs
          see splinepy.helpme.visualize.show

        Returns
        -------
        spline_showable: dict
        """
        return _visualize.show(self, return_showable=True, **kwargs)

    def show(self, **kwargs):
        return _visualize.show(self, **kwargs)

    def boundaries_from_continuity(
        self,
        nthreads=None,
        tolerance=None,
    ):
        """
        Starting from a seed position, the splines are propagated until they
        reach a kink (no g1 continuity). This uses the spline boundary
        information and determines the interface information.

        Parameters
        ----------
        nthreads : int
          Number of threads to be used to determine boundaries and aux values
        tolerance : double
          Tolerance between two normal vectors [cos(angle)] to be considered G1

        Returns
        -------
        None
        """
        if tolerance is None:
            tolerance = _settings.TOLERANCE
        if nthreads is None:
            nthreads = _settings.NTHREADS
        b_patches = self.boundary_multipatch(nthreads=nthreads)

        # Pass information to c++ backend
        self._logd("Start propagation of information...")
        n_new_boundaries = _boundaries_from_continuity(
            b_patches.patches,
            b_patches.interfaces,
            self.interfaces,
            tolerance,
            nthreads,
        )
        self._logd(f"{n_new_boundaries} new boundaries were assigned")

    def combine_boundaries(self, mask=None):
        """
        Combines all boundaries that match an id from the mask to a single
        boundary

        Parameters
        ----------
        mask : array-like
          List of boundary ids to be reassigned a new combined ID

        Returns
        -------
        None
        """
        # retrieve all boundary elements
        boundary_ids = _np.isin(-self.interfaces, _np.abs(mask))

        if not boundary_ids.any():
            self._logd(
                "No boundary elements could be identified that match "
                "requirements"
            )
            return None

        self.interfaces[boundary_ids] = _np.min(mask)

    def add_fields(
        self,
        fields,
        field_dim,
        check_compliance=True,
        check_conformity=True,
        nthreads=None,
    ):
        """
        Add fields using lists of splines

        Parameters
        ----------
        fields : list
          Any number of list of splines to represent n-dimensional field with
          equal parametric dimensionality
        field_dim: int
          Dimensionality of the individual fields
        check_compliance : bool (True)
          Check if field list is admissible, by comparing the parametric
          dimensionality of the field entries with the spline list, and compare
          patch sizes
        check_conformity : bool (True)
          Check for conformity between patches and fields by comparing degrees
          and control-mesh-resolutions
        nthreads : int

        Returns
        -------
        None
        """
        if nthreads is None:
            nthreads = _settings.NTHREADS

        super().add_fields(
            fields,
            field_dim,
            check_compliance,
            check_compliance,
            check_conformity,
            check_conformity,
            nthreads,
        )

    @property
    def fields(self):
        """Save fields as individual splines on patches

        Parameters
        ----------
        None

        Returns
        -------
        fields : list<Multipatch>
          List of all field representation in the form of list of splines
        """
        return super().fields()

    def sample(self, resolutions, nthreads=None):
        """
        Uniformly sample along each parametric dimensions from spline.

        Parameters
        -----------
        resolutions: int
        nthreads: int

        Returns
        --------
        results: (math.product(resolutions), dim) np.ndarray
        """

        if not isinstance(resolutions, int) and hasattr(
            resolutions, "__getitem__"
        ):
            self._logd(
                "sample() only supports uniform sample. Taking first entry."
            )
            # for now, just take the first elem
            resolutions = int(resolutions[0])

        self._logd(
            f"Sampling {resolutions ** self.para_dim} points from spline."
        )
        return super().sample(
            resolutions,
            nthreads=_default_if_none(nthreads, _settings.NTHREADS),
            same_parametric_bounds=False,
        )

    def evaluate(self, queries, nthreads=None):
        """
        Evaluate each individual spline at specific parametric positions. To be
        used with caution, as there is no check if the queries are within the
        parametric bounds if settings.CHECK_BOUNDS is set to false.

        Parameters
        -----------
        queries: (n, para_dim) array-like
        nthreads: int

        Returns
        --------
        results: (math.product(resolutions), dim) np.ndarray
        """

        return super().evaluate(
            queries,
            nthreads=_default_if_none(nthreads, _settings.NTHREADS),
        )

    @property
    def extract(self):
        """Return Extractor object to provide extract functionality for
        multiple splines

        Parameters
        -----------
        None

        Returns
        --------
        extractor: Extractor
        """
        return _get_helper(self, "_extractor", _Extractor)



import numpy as np

import splinepy


def test_multi_index(np_rng):
    """
    Test MultiIndex using a round trip of np.ravel_multi_index
    """
    # 4d? can actually do anything
    for d in range(2, 10 + 1):
        resolutions = np_rng.integers(2, 5, size=d)

        multi = splinepy.helpme.multi_index.MultiIndex(resolutions)

        # test individual id query
        n_q = 3  # number of queries
        query = [np_rng.integers(0, r, size=n_q) for r in resolutions]

        # test index round trip
        # note reversed order - alternatively this should also be the same:
        # np.ravel_multi_index(query, resolutions, order="F")
        ref_raveled = np.ravel_multi_index(query[::-1], resolutions[::-1])
        to_test = multi[tuple(query)]  # can't unpack inside []
        assert np.array_equal(ref_raveled, to_test)

        # test first and last slice that's orthogonal to the last dimension
        # first slice
        ref_raveled = np.arange(np.prod(resolutions[:-1]))
        to_test = multi[..., 0]
        assert np.array_equal(ref_raveled, to_test)

        # last slice
        upper = np.prod(resolutions)
        lower = upper - np.prod(resolutions[:-1])
        ref_raveled = np.arange(lower, upper)
        to_test = multi[..., -1]
        assert np.array_equal(ref_raveled, to_test)

        # test first and last slice that's orthogonal to the first dim
        full_ids = np.arange(np.prod(resolutions))

        # first slice
        ref_raveled = full_ids[0 :: resolutions[0]]
        to_test = multi[0, ...]
        assert np.array_equal(ref_raveled, to_test)
        # syntex check - this should hold in np.ndarray.__getitem__ syntax
        to_test = multi[0, :]
        assert np.array_equal(ref_raveled, to_test)

        # last slice
        ref_raveled = full_ids[int(resolutions[0] - 1) :: resolutions[0]]
        to_test = multi[-1, ...]
        assert np.array_equal(ref_raveled, to_test)



import numpy as np

import splinepy

# Simple Microstructure consisting Bezier-splines only
generator = splinepy.Microstructure()
generator.deformation_function = splinepy.Bezier(
    degrees=[2, 1],
    control_points=[[0, 0], [1, 0], [2, -1], [-1, 1], [1, 1], [3, 2]],
)
generator.microtile = [
    splinepy.Bezier(
        degrees=[3], control_points=[[0, 0.5], [0.5, 1], [0.5, 0], [1, 0.5]]
    ),
    splinepy.Bezier(
        degrees=[4],
        control_points=[
            [0.5, 0],
            [0.75, 0.5],
            [0.8, 0.8],
            [0.25, 0.5],
            [0.5, 1],
        ],
    ),
]
generator.tiling = [8, 8]
ms = generator.create(macro_sensitivities=True)


# Plot the ms with some of its fields
for i in range(0, len(ms.fields), 3):
    ms.spline_data["field"] = i
    ms.show_options["arrow_data"] = "field"
    ms.show_options["arrow_data_on"] = np.linspace(0, 1, 5).reshape(-1, 1)
    ms.show_options["arrow_data_scale"] = 0.1
    ctps, dim = divmod(i, ms.dim)
    splinepy.show(
        [f"Derivative wrt C{ctps},{dim}", ms],
        lighting="off",
        control_points=False,
        resolutions=100,
    )

# Visualize all available microtiles
splinepy.microstructure.tiles.show(
    resolutions=7,
    control_points=False,
    knots=True,
    lighting="off",
)

# Parameter spline
para_spline = splinepy.BSpline(
    degrees=[1, 1],
    knot_vectors=[[0, 0, 2, 2], [0, 0, 1, 1]],
    control_points=[[0.25], [0.3], [0.1], [0.25]],
)


def para_function(x):
    return para_spline.evaluate(x)


def para_sens_function(x):
    return splinepy.utils.data.make_matrix(
        *para_spline.basis_and_support(x),
        para_spline.cps.shape[0],
        as_array=True,
    ).reshape(x.shape[0], 1, para_spline.cps.shape[0])


# Parametrized microstructure inner and outer derivatives
generator = splinepy.Microstructure()
generator.deformation_function = splinepy.BSpline(
    degrees=[1, 1],
    knot_vectors=[[0, 0, 1, 2, 2], [0, 0, 1, 1]],
    control_points=[[0, 0], [1, 0], [2, 0], [0, 1], [1, 1], [2, 1]],
)
generator.tiling = [2, 2]
generator.parameter_sensitivity_function = para_sens_function
generator.parametrization_function = para_function
generator.microtile = splinepy.microstructure.tiles.get("Cross2D")
ms = generator.create(macro_sensitivities=True)
len(ms.fields)

# Plot the ms with its fields
for i in range(0, len(ms.fields), 3):
    ms.spline_data["field"] = i
    ms.show_options["arrow_data"] = "field"
    ctps, dim = divmod(i - 4, ms.dim)
    title = (
        f"Derivative with respect to Parameter-Spline Coefficient {i}"
        if i < 4
        else f"Derivative wrt C{ctps},{dim}"
    )
    splinepy.show(
        [title, ms], lighting="off", control_points=False, knots=False
    )


def foo(x):
    return 0.1 * np.sum(x, axis=1).reshape(-1, 1)


# Cube 3D without closing face
generator = splinepy.Microstructure()
generator.microtile = splinepy.microstructure.tiles.get("HollowCube")

generator.parametrization_function = foo
generator.deformation_function = splinepy.Bezier(
    degrees=[1, 1], control_points=[[0, 0], [1, 0], [0, 1], [1, 1]]
).create.extruded(extrusion_vector=[0, 0, 1])
generator.tiling = [3, 3, 2]
generator.show(
    knots=True,
    control_points=False,
    title="3D Cube Microstructure",
    resolutions=3,
    alpha=0.05,
)


para_s = splinepy.BSpline(
    knot_vectors=[
        [0, 0, 1 / 5, 2 / 5, 3 / 5, 4 / 5, 1, 1],
        [0, 0, 1 / 3, 2 / 3, 1, 1],
    ],
    degrees=[1, 1],
    control_points=[
        [0.15, 0.05],
        [0.15, 0.05],
        [0.15, 0.05],
        [0.15, 0.05],
        [0.05, 0.15],
        [0.05, 0.15],
        [0.05, 0.15],
        [0.05, 0.15],
        [0.15, 0.1],
        [0.15, 0.1],
        [0.15, 0.1],
        [0.05, 0.1],
        [0.05, 0.1],
        [0.1, 0.15],
        [0.15, 0.15],
        [0.05, 0.15],
        [0.15, 0.15],
        [0.15, 0.15],
        [0.2, 0.05],
        [0.2, 0.05],
        [0.2, 0.05],
        [0.15, 0.2],
        [0.15, 0.2],
        [0.15, 0.2],
    ],
)


# Test new microstructure
def parameter_function_double_lattice(x):
    """
    Parametrization Function (determines thickness)
    """
    return para_s.evaluate(x)


generator = splinepy.Microstructure()
generator.deformation_function = splinepy.Bezier(
    degrees=[1, 1], control_points=[[0, 0], [2, 0], [0, 1], [2, 1]]
)
generator.microtile = splinepy.microstructure.tiles.get("DoubleLattice")
generator.tiling = [12, 6]
generator.parametrization_function = parameter_function_double_lattice
my_ms = generator.create(contact_length=0.4)
generator.show(
    use_saved=True,
    knots=True,
    control_points=False,
    title="2D Nuttile Parametrized Microstructure",
    contact_length=0.4,
    resolutions=2,
)
splinepy.show(my_ms, knots=True, control_points=False, resolutions=2)


def parametrization_function_nut(x):
    return np.ones((len(x), 1)) * 0.3


# Structure with Nut-tile
generator = splinepy.Microstructure()
generator.deformation_function = splinepy.Bezier(
    degrees=[1, 1], control_points=[[0, 0], [1, 0], [0, 1], [1, 1]]
)
generator.microtile = splinepy.microstructure.tiles.get("HollowOctagon")
generator.tiling = [10, 10]
generator.parametrization_function = parametrization_function_nut
my_ms = generator.create(closing_face="x", contact_length=0.4)
generator.show(
    use_saved=True,
    knots=True,
    control_points=False,
    title="2D Nuttile Parametrized Microstructure",
    contact_length=0.4,
    resolutions=2,
)


# Classical cross tile 2D
para_cross = splinepy.Bezier(
    degrees=[2, 2],
    control_points=[
        [0.05],
        [0.25],
        [0.35],
        [0.25],
        [0.25],
        [0.2],
        [0.25],
        [0.15],
        [0.15],
    ],
)


def parameter_function_cross_2d(x):
    """
    Parametrization Function (determines branch thickness)
    """
    return para_cross.evaluate(x)


generator = splinepy.Microstructure()
generator.deformation_function = splinepy.Bezier(
    degrees=[1, 1], control_points=[[0, 0], [1, 0], [0, 1], [1, 1]]
)
generator.microtile = splinepy.microstructure.tiles.get("Cross2D")
generator.tiling = [5, 5]
generator.parametrization_function = parameter_function_cross_2d
ms = generator.create(closing_face="x", center_expansion=1.3)
generator.show(
    use_saved=True,
    knots=True,
    control_points=False,
    title="2D Crosstile Parametrized Microstructure",
)

# Micro tile with three bezier lines
generator = splinepy.Microstructure()
generator.deformation_function = splinepy.Bezier(
    degrees=[1, 1], control_points=[[0, 0], [1, 0], [0, 1], [1, 1]]
).create.extruded(extrusion_vector=[0, 0, 1])
generator.microtile = [
    splinepy.Bezier(
        degrees=[3],
        control_points=[
            [0, 0.5, 0.5],
            [0.5, 1, 0.5],
            [0.5, 0, 0.5],
            [1, 0.5, 0.5],
        ],
    ),
    splinepy.Bezier(
        degrees=[3],
        control_points=[
            [0.5, 0.5, 0.0],
            [0.5, 0, 0.5],
            [0.5, 1.0, 0.5],
            [0.5, 0.5, 1.0],
        ],
    ),
    splinepy.Bezier(
        degrees=[3],
        control_points=[
            [0.5, 0, 0.5],
            [1, 0.5, 0.5],
            [0, 0.5, 0.5],
            [0.5, 1, 0.5],
        ],
    ),
]
generator.tiling = [1, 2, 3]
generator.show(
    knots=False, control_points=False, title="3D Lattice Microstructure"
)

# Cross tile 3D in a microstructure
generator = splinepy.Microstructure()
generator.deformation_function = splinepy.Bezier(
    degrees=[1, 1], control_points=[[0, 0], [1, 0], [0, 1], [1, 1]]
).create.extruded(extrusion_vector=[0, 0, 1])
generator.microtile = splinepy.microstructure.tiles.get("Cross3D")
generator.tiling = [2, 2, 3]
generator.show(
    control_points=False, resolutions=2, title="3D Crosstile Microstructure"
)

# Non-uniform tiling
generator = splinepy.Microstructure()
generator.deformation_function = splinepy.BSpline(
    degrees=[2, 1],
    control_points=[
        [0, 0],
        [0.1, 0],
        [0.2, 0],
        [1, 0],
        [0, 1],
        [0.1, 1],
        [0.2, 2],
        [1, 3],
    ],
    knot_vectors=[[0, 0, 0, 0.15, 1, 1, 1], [0, 0, 1, 1]],
)
generator.microtile = [
    splinepy.Bezier(
        degrees=[3], control_points=[[0, 0.5], [0.5, 1], [0.5, 0], [1, 0.5]]
    ),
    splinepy.Bezier(
        degrees=[3], control_points=[[0.5, 0], [1, 0.5], [0, 0.5], [0.5, 1]]
    ),
]
generator.tiling = [5, 1]
generator.show(
    knot_span_wise=False,
    control_points=False,
    resolutions=20,
    title="2D Lattice with global tiling",
)


def foo(x):
    """
    Parametrization Function (determines thickness)
    """
    return (x[:, 0] * 0.05 + x[:, 1] * 0.05 + x[:, 2] * 0.1 + 0.1).reshape(
        -1, 1
    )


# A Parametrized microstructure and its respective inverse structure
generator = splinepy.Microstructure()
generator.deformation_function = splinepy.Bezier(
    degrees=[1, 1], control_points=[[0, 0], [1, 0], [0, 1], [1, 1]]
).create.extruded(extrusion_vector=[0, 0, 1])
generator.microtile = splinepy.microstructure.tiles.get("InverseCross3D")
generator.tiling = [3, 3, 5]
generator.parametrization_function = foo

inverse_microstructure = generator.create(
    closing_face="z", seperator_distance=0.4, center_expansion=1.3
)

# Plot the results
_, showables_inverse = generator.show(
    closing_face="z",
    seperator_distance=0.4,
    center_expansion=1.3,
    title="Parametrized Inverse Microstructure",
    control_points=False,
    knots=True,
    return_showable_list=True,
    resolutions=5,
    c="hotpink",
)

# Corresponding Structure
generator.microtile = splinepy.microstructure.tiles.get("Cross3D")
microstructure = generator.create(
    closing_face="z", seperator_distance=0.4, center_expansion=1.3
)
_, showables = generator.show(
    closing_face="z",
    center_expansion=1.3,
    title="Parametrized Microstructure",
    control_points=False,
    knots=True,
    return_showable_list=True,
    resolutions=5,
)

splinepy.show(
    [*showables_inverse, *showables],
    title="Parametrized Microstructure and its inverse",
)



import numpy as _np

try:
    import scipy as _scipy

    has_scipy = True
except ImportError:
    has_scipy = False


from splinepy import settings as _settings
from splinepy import spline as _spline
from splinepy import splinepy_core as _splinepy_core
from splinepy import utils as _utils


class BSplineBase(_spline.Spline):
    r"""BSpline base. Contains extra operations that are only available for
    bspline families, i.e., :class:`.BSpline` and :class:`.NURBS`.

    Passes all arguments to :code:`super.__init__()`, see :class:`.Spline`.

    All B-Spline-like splines feature knot vectors for each parametric
    dimension. A knot vector can be considered as a set of non-decreasing
    parametric coordinates. If coordinates in the parametric dimension are
    denoted by :math:`u\in\mathbb{R}^{N_{param}}`, then the knot vector
    associated with this dimension contains values

    .. math::
            \Theta_u = \{u_1,u_2,...,u_{l+p+1}\}

    such that

    .. math::
            u_1 \leq u_2 \leq ... \leq u_{l+p+1}

    :math:`u_i` is referred to as the :math:`i`-th *knot*. Intervals of the
    type :math:`[u_i,u_{i+1}]` are known as *knot spans* or *elements*.
    As we can see, the number of knot entries, the number of control points,
    and the polynomial degree of the spline are related to each other by the
    equation :math:`\# knots=l+p+1`. Here, :math:`l` denotes the number of
    control points along a parametric dimension and :math:`p` the polynomial
    order of the spline.
    In the special case of a so-called *open knot vector*, the first and the
    last entry of the knot vector are repeated :math:`p+1` times, that is

    .. math::
          u_1 = ... = u_{p+1} < u_{p+2} < ... < u_l < u_{l+1} = ... = u_{l+p+1}

    Splines with open knot vectors have the property that they exactly
    interpolate the first and the last control point in this parametric
    dimension.

    Based on the knot vectors, the mono-variate basis functions are defined
    recursively via the Cox-de-Boor recursion formula for each parametric
    dimension:

    .. math::
            N^{i;0}(u) &= \begin{cases}
                1 & u\in [u_i,u_{i+1}] \\\\
                0 & \mathrm{else} \end{cases} \\\\
            N^{i;p}(u) &= \frac{u-u_i}{u_{i+p}-u_i} N^{i;p-1}(u)
                + \frac{u_{i+p+1}-u}{u_{i+p+1}-u_{i+1}}N^{i+1;p-1}(u)

    For explanations on how to construct splines using these
    basis functions and usage examples, we refer to the documentation of the
    classes :class:`.BSpline` and :class:`.NURBS`.
    """

    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def insert_knots(self, parametric_dimension, knots):
        """
        Inserts knots.

        Parameters
        -----------
        parametric_dimension: int
          parametric axis, where knots are to be inserted
        knots: list or float
          list of new knots to be inserted into the parametric domain

        Returns
        --------
        inserted: list
          List of bool. True if the knots are inserted. Otherwise, False
        """
        if _settings.CHECK_BOUNDS:
            self.check.clamped_knot_vectors(warning=True)

        if parametric_dimension >= self.para_dim:
            raise ValueError("Invalid parametric dimension to insert knots.")

        if isinstance(knots, float):
            knots = [knots]

        if len(knots) == 0:
            self._logd("Requesting knot insertion for 0 knots, no computation")
            return []

        if max(knots) > max(self.knot_vectors[parametric_dimension]):
            raise ValueError(
                "One of the query knots not in valid knot range. (Too big)"
            )

        if min(knots) < min(self.knot_vectors[parametric_dimension]):
            raise ValueError(
                "One of the query knots not in valid knot range. "
                "(Too small)"
            )

        inserted = _splinepy_core.insert_knots(
            self, parametric_dimension, _utils.data.enforce_contiguous(knots)
        )

        self._logd(f"Inserted {len(knots)} knot(s).")

        self._data = {}

        return inserted

    def uniform_refine(self, para_dims=None, n_knots=1):
        """
        Uniformly refines all knot-spans in given direction(s) by given
        degree(s). By default, every dimension will be refined with 1 knot
        per knot span.

        Parameters
        ----------
        para_dims : int or list
            list of parametric dimensions to be refined (default None -> all)
        n_knots : int or list
            number of new knots per knot span

        Returns
        --------
        None
        """
        # if no para_dim is given - assume that each dimension is refined
        if para_dims is None:
            para_dims = range(self.para_dim)

        # if an integer is given, make it a list
        elif isinstance(para_dims, int):
            para_dims = [para_dims]

        # if an integer is given, assume that each given dimension should be
        # equally refined - make it a list and apply it to all para_dims
        if isinstance(n_knots, int):
            n_knots = [n_knots] * len(para_dims)

        # check if the lengths of both lists are equal
        if len(n_knots) != len(para_dims):
            raise ValueError(
                f"Size mismatch between para_dims ({para_dims}) "
                f"and n_knots ({n_knots})."
            )

        def determine_new_knots(kv_unique, n_knots):
            if n_knots == 0:
                return []
            kv_diffs = _np.diff(kv_unique) / (n_knots + 1)
            new_knots = (
                kv_diffs.reshape(-1, 1)
                @ _np.arange(1, n_knots + 1).reshape(1, -1)
                + kv_unique[:-1].reshape(-1, 1)
            ).ravel()
            return new_knots

        # determine new knots for each para_dim and insert the knots
        for para_dim, n_k in zip(para_dims, n_knots):
            new_knots = determine_new_knots(
                # recompute unique to allow duplicating para_dims.
                kv_unique=self.unique_knots[para_dim],
                n_knots=n_k,
            )
            self.insert_knots(para_dim, new_knots)

    def knot_insertion_matrix(
        self, parametric_dimension=None, knots=None, beziers=False
    ):
        """
        Returns knot insertion matrix for a given set of knots in a specific
        parametric domain, if bezier flag is set, returns matrix, that creates
        C^(-1) spline spaces.

        Matrix can be used to multiply old control points in multi-query
        scenarios. It describes the relation between the old and new control
        points

        .. math::
          c_{new}^i = A c_{old}^i

        Usage:

        .. code-block:: python

            matrix = spline.knot_insertion_matrix(0, [0.1, 0.2])
            spline_copy = spline.copy()
            spline_copy.insert_knots(0, [0.1, 0.2])
            np.allclose(
                spline_copy.control_points, matrix @ spline.control_points
            )

        Parameters
        ----------
        parametric_dimension : int
          parametric dimension along which knots are to be inserted
        knots : array-like
          list of new knots
        beziers : bool (optional)
          if set, all other arguments are ignored. Return matrix will represent
          bezier extraction operation

        Returns
        -------
        matrix : array-like
          Matrix type (scipy sparse if available, else returns full matrix in
          numpy format). Matrix that represents knot insertion. See knot
          insertion for more details
        """
        if _settings.CHECK_BOUNDS:
            self.check.clamped_knot_vectors(warning=True)

        if beziers:
            indices, data = _splinepy_core.bezier_extraction_matrix(
                self,
                _settings.TOLERANCE,
            )

            # Indices represents the relevant ctps whereas data is a list of
            # all matrix information
            # 1. Create a list of matrices
            matrices = []
            for m_data in data:
                if has_scipy:
                    matrix = _scipy.sparse.csr_matrix(
                        m_data[0], shape=m_data[1]
                    )
                else:
                    matrix = _np.zeros(m_data[1])
                    matrix[m_data[0][1][0], m_data[0][1][1]] = m_data[0][0]
                matrices.append(matrix)
            #  2. Create global matrix
            matrix = matrices[0]
            for n_mat in matrices[1:]:
                matrix = n_mat @ matrix

            # 3. Extract Bezier Extraction matrices
            matrices = [matrix[ids, :] for ids in indices]
            return matrices

        data = _splinepy_core.global_knot_insertion_matrix(
            self.knot_vectors,
            self.degrees,
            parametric_dimension,
            _utils.data.enforce_contiguous(knots, dtype="float64"),
            _settings.TOLERANCE,
        )
        if has_scipy:
            matrix = _scipy.sparse.csr_matrix(data[0], shape=data[1])
        else:
            matrix = _np.zeros(data[1])
            matrix[data[0][1][0], data[0][1][1]] = data[0][0]

        return matrix

    def remove_knots(self, parametric_dimension, knots, tolerance=None):
        """
        Tries to removes knots.

        Parameters
        -----------
        parametric_dimension: int
        knots: list or float
        tolerance: float

        Returns
        --------
        removed: list<bool>
          List of bool. True if the knots are Removed. Otherwise, False
        """
        if _settings.CHECK_BOUNDS:
            self.check.clamped_knot_vectors(warning=True)

        if parametric_dimension >= self.para_dim:
            raise ValueError("Invalid parametric dimension to remove knots.")

        if isinstance(knots, float):
            knots = [knots]

        if len(knots) == 0:
            self._logd("Requesting knot removal of 0 knots, no computation")
            return []

        if max(knots) > max(self.knot_vectors[parametric_dimension]):
            raise ValueError(
                "One of the query knots not in valid knot range. (Too big)"
            )

        if min(knots) < min(self.knot_vectors[parametric_dimension]):
            raise ValueError(
                "One of the query knots not in valid knot range. "
                "(Too small)"
            )

        removed = _splinepy_core.remove_knots(
            self,
            parametric_dimension,
            _utils.data.enforce_contiguous(knots),
            tolerance=_spline._default_if_none(tolerance, _settings.TOLERANCE),
        )

        if any(removed):
            self._data = {}

        self._logd(f"Tried to remove {len(knots)} knot(s).")
        self._logd(f"Actually removed {sum(removed)} knot(s).")

        return removed

    def normalize_knot_vectors(self):
        """
        Sets all knot vectors into a range of [0,1], if applicable

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        if not _splinepy_core.has_core(self):
            raise ValueError(
                "spline is not fully initialized."
                "Please, first initialize spline before normalize_knot_vectors."
            )

        for kv in self.knot_vectors:
            if isinstance(kv, _splinepy_core.KnotVector):
                kv.scale(0, 1)

    def extract_bezier_patches(self):
        """
        Extract all knot spans as Bezier patches to perform further operations
        such as compositions and multiplications

        Parameters
        ----------
        None

        Returns
        -------
        extracted Beziers : list
        """
        if _settings.CHECK_BOUNDS:
            self.check.clamped_knot_vectors(warning=True)

        # this core call makes copy of self and extracts bezier patches.
        return _splinepy_core.extract_bezier_patches(self)


class BSpline(BSplineBase):
    r"""
    BSpline.

    Passes all arguments to :code:`super.__init__()`, see
    :class:`.BSplineBase`.

    With the help of the mono-variate basis functions introduced from the
    B-Spline introduction in :class:`.BSplineBase`, we can now construct
    B-Splines.

    .. note::
        For simplicity, we restrict ourselves to the three most common types
        of splines in the following, namely curves, surfaces and volumes,
        although :code:`splinepy` also supports higher dimensions, see
        the documentation of :class:`.Spline` for more information.

    1. A B-Spline of degree :math:`p` with control points
    :math:`P^i\in\mathbb{R}^{N_{phys}}` and a one-dimensional parameter space
    (:math:`N_{param}=1`) corresponds to a line embedded into the physical
    space:

    .. math::
            C(u) = \sum_{i=1}^{l} N^{i;p}(u) P^i

    2. A B-Spline of degrees :math:`p,q` with control points
    :math:`P^{i,j} \in \mathbb{R}^{N_{phys}}` and a two-dimensional parameter
    space (:math:`N_{param}=2`) corresponds to a surface, embedded into the
    physical space:

    .. math::
            S(u,v) = \sum_{i=1}^{l} \sum_{j=1}^{m} N^{i;p}(u) N^{j;q}(v)
                P^{i,j}

    Due to the tensor-product nature of the B-Spline basis functions, this is
    often rewritten in terms of multi-variate basis functions

    .. math::
            \tilde{N}_{i,j;p,q}(u,v) := N^{i;p}(u) N^{j;q}(v)

    3. A B-Spline of degrees :math:`p,q,r` with control points
    :math:`P^{i,j,k} \in \mathbb{R}^{N_{phys}}` and a three-dimensional
    parameter space (:math:`N_{param}=3`) corresponds to a volume, embedded
    into the physical space:

    .. math::
            V(u,v,w) = \sum_{i=1}^{l} \sum_{j=1}^{m} \sum_{k=1}^{n}
                N^{i;p}(u) N^{j;q}(v) N^{k;r}(w) P^{i,j,k}

    Here, we can introduce the multi-variate basis functions

    .. math::
            \tilde{N}^{i,j,k;p,q,r}(u,v,w) := N^{i;p}(u) N^{j;q}(v) N^{k;r}(w)

    Higher-dimensional instances are constructed accordingly.

    **Usage**:

    .. code-block:: python

        bspline_volume = splinepy.BSpline(
            degrees=[1, 1, 1],
            knot_vectors=[
                [0.0, 0.0, 1.0, 1.0],
                [0.0, 0.0, 1.0, 1.0],
                [0.0, 0.0, 1.0, 1.0],
            ],
            control_points=[
                [0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [1.0, 1.0, 0.0],
                [0.0, -1.0, 1.0],
                [1.0, 0.0, 1.0],
                [-1.0, 1.0, 2.0],
                [2.0, 2.0, 2.0],
            ],
        )

    Parameters
    ----------
    degrees: (para_dim) array-like
    knot_vectors: (para_dim) list
        list of array-like
    control_points: (n, dim) array-like
    spline: Spline

    Returns
    -------
    None
    """

    __slots__ = ("_fitting_queries",)

    def __init__(
        self,
        degrees=None,
        knot_vectors=None,
        control_points=None,
        spline=None,
    ):
        super().__init__(
            spline=spline,
            degrees=degrees,
            knot_vectors=knot_vectors,
            control_points=control_points,
        )

    @property
    def bspline(self):
        return self.copy(saved_data=False)

    @property
    def nurbs(self):
        """
        Returns NURBS version of current BSpline by defining all the weights as
        1.

        Parameters
        -----------
        None

        Returns
        --------
        same_nurbs: NURBS
        """
        same_nurbs = _settings.NAME_TO_TYPE["NURBS"](
            **self.todict(),
            weights=_np.ones(self.control_points.shape[0], dtype=_np.float64),
        )

        return same_nurbs



"""Handles the processing of markdown files.

This script is used to process the markdown files in the project that are also
reused in the documentation. The script will replace the relative links in the
markdown files with relative links that are correct when the documentation is
built.

The paths change since the documentation is built from the docs folder and not
from the root of the project.

Author: Clemens Fricke
"""

import os
import pathlib
import re
import warnings

# Path to this file.
file_path = os.path.abspath(os.path.dirname(__file__))
original_cwd = os.getcwd()
repo_root = str(pathlib.Path(__file__).resolve()).split("docs")[0]
os.chdir(repo_root)


def get_markdown_links(line: str) -> str:
    """Get the markdown links from a string.

    Args:
        line (str): text.

    Returns:
        str: Markdown links.
    """
    possible = re.findall(r"\[(.*?)\]\((.*?)\)", line)
    return possible if possible else ""


def get_special_links(line: str) -> list[tuple[str, str]]:
    """Get the special links from a string.

    Args:
        line (str): text.

    Returns:
        List[tuple[str, str]]: Special links.
    """

    possible = [
        x[::-1]
        for x in re.findall(r"<img src=\"(.*?)\".*title=\"(.*?)\">", line)
    ]
    return possible if possible else ""


def get_github_path_from(link):
    """Substitute the path to the github repository.

    This will expand the link to the github repository. This is used to create
    pages that are independent of the documentation. Like for the long
    description on PyPI. Normally we try to use relative links in the files
    to guarantee that the documentation is also available offline. But for
    the long description on PyPI we need to use absolute links to an online
    repository like the github raw files.

    Args:
        link (_type_): _description_

    Returns:
        _type_: _description_
    """
    return str(pathlib.Path(link).resolve()).replace(
        repo_root, "https://raw.githubusercontent.com/tataratat/splinepy/main/"
    )


def get_common_parent(path1: str, path2: str) -> str:
    """Get the common parent of two paths.

    Args:
        path1 (str): Path 1.
        path2 (str): Path 2.

    Returns:
        str: Common parent path.
    """
    path1 = pathlib.Path(path1).resolve().parent
    path2 = pathlib.Path(path2).resolve()
    steps_back = 0
    while path1.is_relative_to(path2) is False:
        path2 = path2.parent
        steps_back += 1
    return path2, steps_back


# Folder to save the processed markdown files to.
folder_to_save_to = os.path.join(repo_root, "docs/md/")

# List of markdown files that are used in the documentation.
markdown_files = [
    pathlib.Path("README.md").resolve(),
    pathlib.Path("CONTRIBUTING.md").resolve(),
    pathlib.Path("docs/markdown/spline_intro.md").resolve(),
    pathlib.Path("docs/markdown/spline_plotting.md").resolve(),
]

# List of explicit link substitutions.
link_substitutions = {
    "docs/markdown/spline_plotting.md": "spline_intro.html#creating-and-plotting-splines"  # noqa E501
}


def process_file(
    file: str, relative_links: bool = True, return_content: bool = False
):
    """Process a markdown file.

    This function will process a markdown file. It will replace all relative
    links with links that are correct when the documentation is built.
    If relative_links is True, the links will be relative to the documentation
    folder. If relative_links is False, the links will be absolute links to
    the github repository.

    Args:
        file (str): Path to the markdown file.
        relative_links (bool, optional):
            Generate relative links. Defaults to True.
        return_content (bool, optional):
            Return the content instead of saving it. Defaults to False.

    Returns:
        str: Content of the markdown file. Only if return_content is True.
    """
    # read in the content of the markdown file
    with open(file) as f:
        content = f.read()
    os.chdir(os.path.dirname(file)) if os.path.dirname(file) else None
    # get all links from the markdown file
    links = get_markdown_links(content)

    for item in links:
        if not item[1].strip():
            warnings.warn(
                f"Empty link in `{file}`. Link name `{item[0]}` link path "
                f"`{item[1]}`. Will ignore link.",
                stacklevel=3,
            )
            continue
        if item[1].startswith(("http", "#")):  # skip http links and anchors
            if "badge" in item[1]:
                continue
            content = content.replace(
                f"[{item[0]}]({item[1]})",
                f"<a href='{item[1]}'>{item[0]}</a>",
            )
            continue
        elif item[1] in link_substitutions:
            if relative_links:
                content = content.replace(
                    f"[{item[0]}]({item[1]})",
                    f"<a href='{link_substitutions[item[1]]}'>{item[0]}</a>",
                )
                continue
            else:
                content = content.replace(
                    f"[{item[0]}]({item[1]})",
                    "See documentation for examples.",
                )
        elif not relative_links:  # generate links to github repo
            new_path = get_github_path_from(pathlib.Path(item[1]).resolve())
        else:  # generate relative links
            common_sub_path, steps_back = get_common_parent(
                item[1], folder_to_save_to
            )
            new_path = "../" * steps_back + str(
                pathlib.Path(item[1]).resolve().relative_to(common_sub_path)
            )
        content = content.replace(item[1], str(new_path))

    # super special links (just special images) that the sphinx markdown
    # parser won't correctly handle since they are in html tags.
    special_links = get_special_links(content)
    for item in special_links:
        if not item[0].strip():
            warnings.warn(
                f"Empty link in `{file}`. Link name `{item[1]}` link path "
                f"`{item[0]}`. Will ignore link.",
                stacklevel=3,
            )
            continue
        if item[0].startswith("http"):  # skip http links and anchors
            continue
        else:
            # just link to static folder in docs
            new_path = "_static/" + str(pathlib.Path(item[1]).name)
            content = content.replace(item[1], str(new_path))

    os.chdir(original_cwd)

    if return_content:
        return content

    with open(
        os.path.join(folder_to_save_to, os.path.basename(file)), "w"
    ) as f:
        f.write(content)


if __name__ == "__main__":
    os.chdir(repo_root)
    os.makedirs(folder_to_save_to, exist_ok=True)
    # Process all markdown files
    for file in markdown_files:
        process_file(file)



import numpy as np

import splinepy


def test_insert_knot(np_rng, are_items_close, bspline_2p2d, nurbs_2p2d):
    """Test the knot insertion function (.insert_knot())."""

    # creating a reference copy of the splines
    bspline_original = bspline_2p2d.copy()
    nurbs_original = nurbs_2p2d.copy()

    # reference solutions
    bspline_ref_kv = [
        [0, 0, 0, 0.2, 0.5, 0.7, 1, 1, 1],
        [0, 0, 0, 0.2, 0.5, 0.7, 1, 1, 1],
    ]
    nurbs_ref_kv = [
        [0, 0, 0, 0.2, 0.5, 0.7, 1, 1, 1],
        [0, 0, 0.2, 0.5, 0.7, 1, 1],
    ]

    # insert knots
    bspline_2p2d.insert_knots(0, [0.2, 0.7])
    bspline_2p2d.insert_knots(1, [0.2, 0.5, 0.7])

    nurbs_2p2d.insert_knots(0, [0.2, 0.5, 0.7])
    nurbs_2p2d.insert_knots(1, [0.2, 0.5, 0.7])

    # test knot_vectors
    assert are_items_close(bspline_2p2d.knot_vectors, bspline_ref_kv)

    assert are_items_close(nurbs_2p2d.knot_vectors, nurbs_ref_kv)

    # use random query points
    q2D = np_rng.random((10, 2))

    # test evaluation
    assert np.allclose(
        bspline_2p2d.evaluate(q2D), bspline_original.evaluate(q2D)
    )

    assert np.allclose(nurbs_2p2d.evaluate(q2D), nurbs_original.evaluate(q2D))


def test_insert_knot_status_return(np_rng):
    # create 3d box and loop for bspline families
    box = splinepy.helpme.create.box(1, 1, 1)
    for s in [box.bspline, box.nurbs]:
        s.elevate_degrees([0, 1, 2])
        degrees = s.degrees

        # create random knots to insert
        knots = np_rng.random(5)
        for i in range(s.para_dim):
            # repeat knots for c^-1 repetition - they should all be possible
            repeat = degrees[i] + 1
            repeated = np.repeat(knots, repeat)

            # we create reference
            # last n_knots should fail
            ref = np.ones(repeated.size + knots.size, dtype=bool)
            ref[-knots.size :] = False

            # schuffle
            np_rng.shuffle(repeated)
            repeated = np.append(repeated, knots)

            successful = s.insert_knots(i, repeated)
            assert all(successful == ref)


def test_insert_knot_with_matrix(np_rng, bspline_2p2d, nurbs_2p2d):
    """Test the knot insertion function (.insert_knot())."""

    # creating a reference copy of the splines
    bspline_original = bspline_2p2d.copy()
    nurbs_original = nurbs_2p2d.copy()

    # BSpline Data
    b_spline_knots_0 = np_rng.random(8)
    b_spline_knots_1 = np_rng.random(9)
    matrix_bspline = bspline_2p2d.knot_insertion_matrix(0, b_spline_knots_0)
    bspline_2p2d.insert_knots(
        0,
        b_spline_knots_0,
    )
    matrix_bspline = (
        bspline_2p2d.knot_insertion_matrix(1, b_spline_knots_1)
        @ matrix_bspline
    )
    bspline_2p2d.insert_knots(1, b_spline_knots_1)

    # NURBS Data
    nurbs_knots_0 = np_rng.random(10)
    nurbs_knots_1 = np_rng.random(11)
    matrix_nurbs = nurbs_2p2d.knot_insertion_matrix(0, nurbs_knots_0)
    nurbs_2p2d.insert_knots(0, nurbs_knots_0)
    matrix_nurbs = (
        nurbs_2p2d.knot_insertion_matrix(1, nurbs_knots_1) @ matrix_nurbs
    )
    nurbs_2p2d.insert_knots(1, nurbs_knots_1)

    # use random query points
    q2D = np_rng.random((50, 2))

    # test evaluation
    assert np.allclose(
        bspline_2p2d.evaluate(q2D), bspline_original.evaluate(q2D)
    )

    assert np.allclose(nurbs_2p2d.evaluate(q2D), nurbs_original.evaluate(q2D))

    # Test control points and weights
    assert np.allclose(
        bspline_2p2d.control_points,
        matrix_bspline @ bspline_original.control_points,
    )

    assert np.allclose(
        nurbs_2p2d.weights, matrix_nurbs @ nurbs_original.weights
    )

    assert np.allclose(
        nurbs_2p2d.control_points,
        matrix_nurbs
        @ (nurbs_original.weights * nurbs_original.control_points)
        / nurbs_2p2d.weights,
    )


def test_remove_knot(np_rng, are_items_close, bspline_2p2d, nurbs_2p2d):
    """Test the function .remove_knots.
    This test also depends on the function .insert_knots!"""

    # creating a reference copy of the splines
    bspline_original = bspline_2p2d.copy()
    nurbs_original = nurbs_2p2d.copy()

    # insert and remove knots
    bspline_2p2d.insert_knots(0, [0.2, 0.7])
    bspline_2p2d.insert_knots(1, [0.2, 0.5, 0.7])
    bspline_2p2d.remove_knots(0, [0.2, 0.7])
    bspline_2p2d.remove_knots(1, [0.2, 0.5, 0.7])

    nurbs_2p2d.insert_knots(0, [0.2, 0.5, 0.7])
    nurbs_2p2d.insert_knots(1, [0.2, 0.5, 0.7])
    nurbs_2p2d.remove_knots(0, [0.2, 0.5, 0.7])
    nurbs_2p2d.remove_knots(1, [0.2, 0.5, 0.7])

    # test knot_vectors
    assert are_items_close(
        bspline_2p2d.knot_vectors, bspline_original.knot_vectors
    )

    assert are_items_close(
        nurbs_2p2d.knot_vectors, nurbs_original.knot_vectors
    )

    # use random query points
    q2D = np_rng.random((10, 2))

    # test evaluation
    assert np.allclose(
        bspline_2p2d.evaluate(q2D), bspline_original.evaluate(q2D)
    )

    assert np.allclose(nurbs_2p2d.evaluate(q2D), nurbs_original.evaluate(q2D))


def test_uniform_refine():
    """test uniform refine from bspline families by
    checking resulting number of elements"""

    # create single element bspline
    b = splinepy.helpme.create.box(1, 2, 3).bspline
    n = b.nurbs

    for spline in [b, n]:
        # check default - one subdivision
        spline.uniform_refine()
        unique_knots1 = spline.unique_knots
        n_elem1 = 1
        for uk in unique_knots1:
            n_elem1 *= len(uk) - 1

        assert n_elem1 == 2**3

        # add different numbers
        n_knots = [1, 2, 3]
        spline.uniform_refine(n_knots=n_knots)
        unique_knots2 = spline.unique_knots
        n_elem2 = 1
        for uk in unique_knots2:
            n_elem2 *= len(uk) - 1

        # compute what it should be
        n_elem2_ref = 1
        for uk, nk in zip(unique_knots1, n_knots):
            n_elem2_ref *= (len(uk) - 1) * (nk + 1)

        assert n_elem2_ref == n_elem2

        # try only one axis
        n_knots = [2]
        p_dim = [0]
        spline.uniform_refine(p_dim, n_knots)
        unique_knots3 = spline.unique_knots
        n_elem3 = 1
        for uk in unique_knots3:
            n_elem3 *= len(uk) - 1
        assert n_elem2 * (n_knots[0] + 1) == n_elem3



import numpy as _np

from splinepy import settings as _settings
from splinepy.helpme.multi_index import MultiIndex
from splinepy.utils import log as _log
from splinepy.utils.data import has_scipy as _has_scipy
from splinepy.utils.data import make_matrix as _make_matrix

if _has_scipy:
    from scipy.sparse.linalg import spsolve as _spsolve


def parameterize(fitting_points, size, centripetal):
    """
    Parameterizes given fitting_points later used
    in interpolation/approximation.
    Values are then used to build adequate knot vectors.

    Parameters
    ----------
    fitting_points:(n, dim) array-like
      points to be interpolated/approximated
    size: array-like
      number of points per parametric dimension
    centripetal: bool
      if True -> centripetal parameterization will be used

    Returns
    -------
    parametric_coordinates: list<np.ndarray>
      array containing parameterization values per parametric dimension
    """

    def parametrize_to_line(reorganized_queries, axis, centripetal):
        """parameterize_line all at once.
        Expects with-MultiIndex-Reorganized points"""
        n_points = reorganized_queries.shape[axis]
        u_k = _np.empty(n_points)

        # norm from axis -1 as last axis is dim.
        chord_lengths = _np.linalg.norm(
            _np.diff(reorganized_queries, axis=axis), axis=-1
        )

        if centripetal:
            chord_lengths = _np.sqrt(chord_lengths)

        # indices for total chord_reshapes
        chord_lengths_shape = list(chord_lengths.shape)
        chord_lengths_shape[axis] = 1

        total_chord_length = _np.sum(chord_lengths, axis=axis).reshape(
            *chord_lengths_shape
        )
        chord_lengths /= total_chord_length

        # Compute cumulated sum
        if chord_lengths.ndim == 1:
            u_k[1:] = _np.cumsum(chord_lengths, axis=axis)
        else:
            # Indices for slices
            mean_slicer = list(range(fitting_points.ndim))
            mean_slicer.pop(axis)
            mean_slicer = tuple(mean_slicer)
            # Higher dimensional (surface ++)
            u_k[1:] = _np.mean(
                _np.cumsum(chord_lengths, axis=axis),
                axis=mean_slicer,
            )
        u_k[0] = 0.0
        # Numerical inaccuracies where an issue with the last value sometimes
        u_k[-1] = 1.0

        return u_k.reshape(-1, 1)

    mi = MultiIndex(size)

    # Loop over all dimensions and append each para_coords to list
    reorganized = fitting_points[mi._raveled_indices]
    parametric_coordinates = []
    for k in range(len(size)):
        parametric_coordinates.append(
            parametrize_to_line(reorganized, k, centripetal)
        )

    return parametric_coordinates


def compute_knot_vector(degree, n_control_points, u_k, n_fitting_points):
    """
    Computes the knot_vector for a spline.

    Parameters
    ----------
    degree: int
      degree of spline
    n_control_points: int
      number of control points
    u_k: (n, 1) array-like
      array containing parameterization values
    n_fitting_points: int
      number of fitting_points

    Returns
    -------
    knot_vector: (degree + n_control_points + 1, 1) np.ndarray
      array containing knots of spline
    """
    # initialize with m = n + p + 1
    knot_vector = _np.empty(degree + n_control_points + 1)

    knot_vector[: (degree + 1)] = 0.0
    knot_vector[-(degree + 1) :] = 1.0

    # check if interpolation or approximation
    if n_fitting_points == n_control_points:
        # interpolation
        # NURBS Book eq. 9.8 (n_fitting_points = m + 1,
        # n_control_points = n + 1)
        knot_vector[(degree + 1) : -(degree + 1)] = (
            _np.convolve(u_k.ravel(), _np.ones(degree), "valid")[1:-1] / degree
        )

    else:
        # approximation
        # NURBS Book eq. (9.68) (n_fitting_points = m + 1,
        # n_control_points = n + 1)
        d = (n_fitting_points) / (n_control_points - degree)

        u_k_flat = u_k.flat

        # NURBS Book eq. (9.69)
        for j in range(1, n_control_points - degree):
            i = int(j * d)
            alpha = (j * d) - i
            knot_vector[j + degree] = (1 - alpha) * u_k_flat[
                i - 1
            ] + alpha * u_k_flat[i]
    return knot_vector


def solve_for_control_points(
    fitting_points, fitting_spline, queries, interpolate_endpoints
):
    """
    Builds and solves the system to calculate control points.

    Parameters
    ----------
    fitting_points:(m + 1 x dim) array
        points to be interpolated/approximated
    fitting_spline:spline
        spline used for interpolation/approximation
    queries:(n x 1) array
        parametric values where the spline will be evaluated
    interpolate_endpoints:bool
        if True -> endpoints are interpolated

    Returns
    -------
    residual:float
        residual (coefficient_matrix @ control_points - fitting_points)
    """
    # build matrix
    coefficient_matrix = _make_matrix(
        *fitting_spline.basis_and_support(queries),
        fitting_spline.control_points.shape[0],
    )

    if (
        interpolate_endpoints
        and fitting_spline.control_points.shape != fitting_points.shape
    ):
        # reduction of matrix necessary for interpolation of endpoints
        # copy fitting_points to keep original fitting_points unchanged
        rhs = fitting_points.copy()

        fitting_spline.control_points[0] = rhs[0]
        fitting_spline.control_points[-1] = rhs[-1]

        if fitting_spline.control_points.shape[0] == 2:
            # control points equal to endpoints (straight line)
            residual = _np.linalg.norm(
                coefficient_matrix @ fitting_spline.control_points
                - fitting_points
            )
        elif _has_scipy:
            # move known values to RHS
            rhs[1:-1, :] -= (
                coefficient_matrix[1:-1, [0]] * fitting_points[0, :]
                + coefficient_matrix[1:-1, [-1]] * fitting_points[-1, :]
            ).todense()

            # solve system A^TAx = A^Tb
            # (tried using scipy.sparse.linalg.lsqr but had worse performance)
            at = coefficient_matrix[1:-1, 1:-1].transpose(copy=True)
            fitting_spline.control_points[1:-1] = _spsolve(
                at @ coefficient_matrix[1:-1, 1:-1], at @ rhs[1:-1]
            )
        else:
            # move known values to RHS
            rhs[1:-1, :] += (
                -coefficient_matrix[1:-1, [0]] * fitting_points[0, :]
                - coefficient_matrix[1:-1, [-1]] * fitting_points[-1, :]
            )

            # solve system using lstsq

            fitting_spline.control_points[1:-1] = _np.linalg.lstsq(
                coefficient_matrix[1:-1, 1:-1], rhs[1:-1]
            )[0]

    # interpolation or approximation with original system
    elif _has_scipy:
        at = coefficient_matrix.transpose(copy=True)
        fitting_spline.control_points = _spsolve(
            at @ coefficient_matrix, at @ fitting_points
        )

    else:
        fitting_spline.control_points = _np.linalg.lstsq(
            coefficient_matrix, fitting_points
        )[0]

    residual = _np.linalg.norm(
        coefficient_matrix @ fitting_spline.control_points - fitting_points
    )
    return residual


def _validate_specifications(n_query, degree, n_control_points, knot_vector):
    """Given main values of fitting, validates degree and n_control_points.
    If there's mismatch between expected value and given value, values will be
    overwritten."""
    # sanity checks for given values
    if degree is not None and knot_vector is not None:
        expected_ncps = len(knot_vector) - degree - 1
        if n_control_points is not None and n_control_points != expected_ncps:
            _log.error(
                f"n_control_points should be {expected_ncps}. Overwriting."
            )

        n_control_points = expected_ncps

    if knot_vector is not None and n_control_points is not None:
        expected_degree = len(knot_vector) - n_control_points - 1
        if degree is not None and degree != expected_degree:
            _log.error(f"degree should be {expected_degree}. Overwriting.")

        degree = expected_degree

    # we need degree
    if degree is None:
        raise ValueError(
            "Not enough input to determine degree. Please set degree."
        )

    # no n_control point -> same as query and this will be interpolation
    if n_control_points is None:
        _log.debug(
            "n_control_points not specified. "
            "Setting it the same as the number of queries"
        )
        n_control_points = n_query

    # degree check
    if degree >= n_control_points:
        raise ValueError("Given degree must be lower than n_control_points!")

    # check number of fitting_points and control points
    if n_control_points > n_query:
        raise ValueError(
            f"(n_fitting_points({n_query}) >= "
            f"n_control_points({n_control_points})) not satisfied!"
        )

    return degree, n_control_points


def _prepare_default_bspline1pd(
    n_queries, dim, degree, n_control_points, knot_vector, u_k
):
    # validate values. may raise if there's conflict.
    degree, n_control_points = _validate_specifications(
        n_queries, degree, n_control_points, knot_vector
    )

    # empty init of default fitting spline
    fitting_spline = _settings.NAME_TO_TYPE["BSpline"](
        [degree],
        [
            compute_knot_vector(
                degree=degree,
                n_control_points=n_control_points,
                u_k=u_k,
                n_fitting_points=n_queries,
            )
        ],
        _np.empty((n_control_points, dim)),
    )

    return fitting_spline


def curve(
    fitting_points,
    degree=None,
    n_control_points=None,
    knot_vector=None,
    fitting_spline=None,
    associated_queries=None,
    centripetal=True,
    interpolate_endpoints=True,
    verbose_output=False,
):
    """
    Fits a spline with given parameters through given fitting_points.
    Spline will be interpolated if n_control_points = n_fitting_points and
    approximated if not.

    Parameters
    ----------
    fitting_points: (m, dim) array
        points to be interpolated/approximated
    degree: int
        degree of spline
    n_control_points: int
        number of control points
    knot_vector: list
        desired knot vector of spline
    fitting_spline: Spline
        spline used for interpolation/approximation
    associated_queries: (n, 1) np.ndarray
        values where the spline will be evaluated
    centripetal: bool (default = True)
        if True -> centripetal parameterization will be used
    interpolate_endpoints: bool
        if True -> endpoints are interpolated
    verbose_outputs: dict
        returns additional information as dict

    Returns
    -------
    fitted_spline: Spline
        interpolated/approximated spline
    residual: float
        residual (coefficient_matrix @ control_points - fitting_points)
    """
    fitting_points = _np.asanyarray(fitting_points)

    # calculate n_fitting_points due to multiple usage
    n_fitting_points = fitting_points.shape[0]

    # determine evaluation points to build a linear system
    if associated_queries is not None:
        u_k = _np.asanyarray(associated_queries)
        if u_k.shape[1] != 1:
            raise ValueError("Associated queries need to have (-1, 1) shape")

    else:
        u_k = parameterize(
            fitting_points=fitting_points,
            size=[n_fitting_points],
            centripetal=centripetal,
        )[0]

    # create fitting_spline
    if fitting_spline is not None:
        # spline dimension check
        if fitting_spline.para_dim != 1:
            raise ValueError(
                "parametric dimension of fitting_spline must be 1!"
            )

        # some values maybe ignored.
        _log.debug(
            "degrees, n_control_points, or knot_vector maybe ignored as "
            "fitting_spline given."
        )

        # create copy, as we will set control points directly
        fitting_spline = fitting_spline.copy()

    else:
        # create bspline based on specifications
        fitting_spline = _prepare_default_bspline1pd(
            *fitting_points.shape, degree, n_control_points, knot_vector, u_k
        )

    # solve system for control points
    residual = solve_for_control_points(
        fitting_points=fitting_points,
        fitting_spline=fitting_spline,
        queries=u_k,
        interpolate_endpoints=interpolate_endpoints,
    )

    if verbose_output:
        return {
            "knot_vector": fitting_spline.knot_vector,
            "degree": fitting_spline.degrees,
            "control_points": fitting_spline.control_points,
            "fitting_points": fitting_points,
            "associated_queries": u_k,
            "residual": residual,
        }

    return fitting_spline, residual


def surface(
    fitting_points,
    size,
    degrees=None,
    n_control_points=None,
    knot_vectors=None,
    fitting_spline=None,
    associated_queries=None,
    centripetal=True,
    interpolate_endpoints=True,
    verbose_output=False,
):
    """
    Fits a surface spline with given parameters through given fitting_points.
    Spline will be interpolated if
    n_control_points[0]*n_control_points[1] = n_fitting_points and
    approximated if not.

    Parameters
    ----------
    fitting_points: (m + 1 x dim) array
        points to be interpolated/approximated
    degrees: array-like (int int)
        degree of fitted spline in both directions
    n_control_points: array-like (int, int)
        number of control points in each direction
    knot_vectors: list of lists [list list]
        knot vectors of fitted spline in both directions
    fitting_spline: spline
        spline used for interpolation/approximation
        must have parametric dimension of 2
    associated_queries: list of lists [list list]
        values where the splines will be evaluated (both directions!)
        will only be used if knot_vectors or fitting_spline is also given!
    centripetal: bool
        Default is True. if True -> centripetal parameterization will be used
    interpolate_endpoints: bool
        if True -> endpoints are interpolated
    verbose_outputs:
        returns additional information as dict

    Returns
    -------
    fitted_spline: spline
        interpolated/approximated spline
    residual: float
        residual (coefficient_matrix @ control_points - fitting_points)
    """
    fitting_points = _np.asanyarray(fitting_points)
    dim = fitting_points.shape[1]

    # get evaluation locations
    u_k = [None, None]
    if associated_queries is not None and (
        knot_vectors is not None or fitting_spline is not None
    ):
        # check dimensions of queries
        if (
            associated_queries[0].shape[1] != 1
            or associated_queries[1].shape[1] != 1
        ):
            raise ValueError(
                "Associated queries in each direction must have dimension 1!"
            )
        u_k[0] = _np.asanyarray(associated_queries[0])
        u_k[1] = _np.asanyarray(associated_queries[1])

    else:
        u_k = parameterize(
            fitting_points=fitting_points, size=size, centripetal=centripetal
        )

    # initialize list for fitting splines
    # structure is needed due to function calls later
    fitting_splines = [None, None]
    if fitting_spline is None:
        # initialize knot_vectors
        if knot_vectors is None:
            knot_vectors = [None, None]
        if degrees is None:
            degrees = [None, None]
        if n_control_points is None:
            n_control_points = [None, None]

        # create base spline for each direction
        spline_u = _prepare_default_bspline1pd(
            size[0],
            dim,
            degrees[0],
            n_control_points[0],
            knot_vectors[0],
            u_k[0],
        )
        spline_v = _prepare_default_bspline1pd(
            size[1],
            dim,
            degrees[1],
            n_control_points[1],
            knot_vectors[1],
            u_k[1],
        )
        fitting_splines = [spline_u, spline_v]
        fitted_spline = type(spline_u)(
            [*spline_u.ds, *spline_v.ds],
            [*spline_u.kvs, *spline_v.kvs],
            _np.empty((len(spline_u.cps) * len(spline_v.cps), dim)),
        )

    else:
        if fitting_spline.para_dim != 2:
            raise ValueError(
                "If fitting spline is given as one spline, "
                "parametric dimension must be 2!"
            )
        # extract spline in each direction (knot_vectors, weights etc.
        # of original spline), works for all type of splines!
        fitting_splines = fitting_spline.extract.boundaries([2, 0])
        fitted_spline = fitting_spline.copy()

    n_control_points = fitted_spline.control_mesh_resolutions

    # index helpers
    mi_pts = MultiIndex(size)
    mi_interim_cps = MultiIndex((n_control_points[0], size[1]))

    # loop first dim
    # curve fit for every j in n_points_v
    interim_control_points = _np.empty((n_control_points[0] * size[1], dim))
    residual_u = _np.empty(size[1])
    fitted_spline_u = fitting_splines[0]
    for v in range(size[1]):
        # previously fitted spline is used for the other fits
        fitted_spline_u, residual_u[v] = curve(
            fitting_points=fitting_points[mi_pts[:, v]],
            fitting_spline=fitted_spline_u,
            associated_queries=u_k[0],
            centripetal=centripetal,
            interpolate_endpoints=interpolate_endpoints,
        )
        # cps in u direction (later fitted in v direction)
        interim_control_points[mi_interim_cps[:, v]] = (
            fitted_spline_u.control_points
        )

    # loop second dim
    # curve fit for every k in n_control_points_u
    residual_v = _np.empty(n_control_points[0])
    fitted_spline_v = fitting_splines[1]
    for u in range(n_control_points[0]):
        # previously fitted spline is used for the other fits
        fitted_spline_v, residual_v[u] = curve(
            fitting_points=interim_control_points[mi_interim_cps[u, :]],
            fitting_spline=fitted_spline_v,
            associated_queries=u_k[1],
            centripetal=centripetal,
            interpolate_endpoints=interpolate_endpoints,
        )

        interim_control_points[mi_interim_cps[u, : n_control_points[1]]] = (
            fitted_spline_v.control_points
        )

    # copy
    fitted_spline.control_points = interim_control_points[
        mi_interim_cps[: n_control_points[0], : n_control_points[1]]
    ]

    residual = _np.linalg.norm(
        (_np.linalg.norm(residual_u), _np.linalg.norm(residual_v))
    )

    if verbose_output:
        return {
            "knot_vectors": fitted_spline.knot_vector,
            "degrees": fitted_spline.degrees,
            "control_points": fitted_spline.control_points,
            "fitting_points": fitting_points,
            "associated_queries": u_k,
            "residual": residual,
        }
    return fitted_spline, residual



"""
Export splines in custom json format
"""

import base64 as _base64
import json as _json

import numpy as _np

from splinepy import settings as _settings
from splinepy.utils.log import debug as _debug


def load(fname):
    """
    Loads splines from json file. Returns a list of splines

    Parameters
    -----------
    fname: str

    Returns
    --------
    spline_list: list
    """
    # Make Required Properties locally accessible
    from splinepy.spline import RequiredProperties as _RequiredProperties

    # Import data from file into dict format
    with open(fname) as f:
        jsonbz = _json.load(f)

    spline_list = []
    base64encoding = jsonbz["Base64Encoding"]

    for jbz in jsonbz["SplineList"]:
        # Convert Base64 str into np array
        if base64encoding:
            jbz["control_points"] = _np.frombuffer(
                _base64.b64decode(jbz["control_points"].encode("ascii")),
                dtype=_np.float64,
            ).reshape((-1, jbz["dim"]))
            if jbz.get("weights") is not None:
                jbz["weights"] = _np.frombuffer(
                    _base64.b64decode(jbz["weights"].encode("ascii")),
                    dtype=_np.float64,
                ).reshape((-1, 1))

        # Declare Type and get required properties
        req_props = _RequiredProperties.of(jbz["SplineType"])
        data_dict = {}
        for prop in req_props:
            data_dict[prop] = jbz[prop]
        spline_list.append(
            _settings.NAME_TO_TYPE[jbz["SplineType"]](**data_dict)
        )

        # Add show options if any
        show_options = jbz.get("show_options", None)
        if show_options is not None:
            for key, value in show_options.items():
                spline_list[-1].show_options[key] = value

    _debug("Imported " + str(len(spline_list)) + " splines from file.")

    return spline_list


def export(fname, spline_list, list_name=None, base64encoding=False):
    """
    Exports a list of arbitrary splines in json-format

    Parameters
    ----------
    fname : str
      Export Filename
    spline_list: list
      List of arbitrary Spline-Types
    list_name: str
      Default is None and "SplineGroup" will be assigned. Used to define name.
    base64encoding: bool
      Default is False. If True, encodes float type spline properties before
      saving.

    Returns
    -------
    output_dict : dict
      Dictornay data written into file

    """
    # Determine Spline Group
    if list_name is None:
        list_name = "SplineGroup"

    # Ensure spline_list is actually a list
    if not isinstance(spline_list, list):
        spline_list = [spline_list]

    # Number of splines in group
    n_splines = len(spline_list)

    output_dict = {
        "Name": list_name,
        "NumberOfSplines": n_splines,
        "Base64Encoding": base64encoding,
    }

    # Append all splines to a dictionary
    from splinepy.spline import Spline as _Spline

    spline_dictonary_list = []
    for i_spline in spline_list:
        if not issubclass(type(i_spline), _Spline):
            raise ValueError("Unsupported type in list")

        # Create Dictionary
        i_spline_dict = i_spline.todict(tolist=True)
        i_spline_dict["SplineType"] = type(i_spline).__qualname__
        # Append para_dim and dim
        i_spline_dict["dim"] = i_spline.dim
        i_spline_dict["para_dim"] = i_spline.para_dim

        if base64encoding:
            i_spline_dict["control_points"] = _base64.b64encode(
                _np.array(i_spline_dict["control_points"])
            ).decode("utf-8")
            if "weights" in i_spline.required_properties:
                i_spline_dict["weights"] = _base64.b64encode(
                    _np.array(i_spline_dict["weights"])
                ).decode("utf-8")

        i_spline_dict["show_options"] = dict(i_spline.show_options)

        spline_dictonary_list.append(i_spline_dict)

    output_dict["SplineList"] = spline_dictonary_list

    with open(fname, "w") as file_pointer:
        file_pointer.write(_json.dumps(output_dict, indent=4))

    return output_dict



"""Create gus"""

import gustaf as gus
import numpy as np

import splinepy

WITH_PUPILS = False
WITH_DARK_BACKGROUND = False
DARK_WITH_WHITE_OVAL_BACKGROUND = False
WITH_ALPHA_BACKGROUND = False

if __name__ == "__main__":
    if WITH_ALPHA_BACKGROUND:
        import vedo

        vedo.settings.screenshot_transparent_background = 1

    l_eye = splinepy.helpme.create.disk(0.15)
    l_eye.cps += [0.6, 0.2]
    l_eye.show_options["c"] = "black"
    l_eye.show_options["control_points"] = False
    l_eye.show_options["knots"] = False

    r_eye = splinepy.helpme.create.disk(0.15)
    r_eye.cps += [-0.6, 0.2]
    r_eye.show_options["c"] = "black"
    r_eye.show_options["control_points"] = False
    r_eye.show_options["knots"] = False

    x_offset = 0.59
    y_offset = 0.16

    l_pupil = splinepy.helpme.create.disk(0.05)
    l_pupil.cps += [x_offset, y_offset]
    l_pupil.show_options["c"] = "white"
    l_pupil.show_options["control_points"] = False
    l_pupil.show_options["knots"] = False

    r_pupil = splinepy.helpme.create.disk(0.05)
    r_pupil.cps += [-x_offset, y_offset]
    r_pupil.show_options["c"] = "white"
    r_pupil.show_options["control_points"] = False
    r_pupil.show_options["knots"] = False

    upper_lip = splinepy.Bezier(
        [5, 1],
        [
            [-0.8, -0.1],
            [-0.2, -0.4],
            [-0.5, 0.5],
            [0.5, 0.5],
            [0.2, -0.4],
            [0.8, -0.1],
            [-0.75, -0.15],
            [-0.2, -0.55],
            [-0.5, 0.05],
            [0.5, 0.05],
            [0.2, -0.55],
            [0.75, -0.15],
        ],
    )
    upper_lip.show_options["c"] = "orange7"
    upper_lip.show_options["control_points"] = False
    upper_lip.show_options["knots"] = False

    inner_mouth = splinepy.Bezier(
        [5, 1],
        [
            *upper_lip.cps[6:],
            [-0.75, -0.15],
            [-0.6, -0.325],
            [-0.4, -0.525],
            [0.4, -0.525],
            [0.6, -0.325],
            [0.75, -0.15],
        ],
    )
    inner_mouth.show_options["c"] = "orange5"
    inner_mouth.show_options["control_points"] = False
    inner_mouth.show_options["knots"] = False

    lower_lip = splinepy.Bezier(
        [5, 1],
        [
            *inner_mouth.cps[6:],
            [-0.8, -0.1],
            [-0.6, -0.4],
            [-0.4, -0.6],
            [0.4, -0.6],
            [0.6, -0.4],
            [0.8, -0.1],
        ],
    )
    lower_lip.show_options["c"] = "orange7"
    lower_lip.show_options["control_points"] = False
    lower_lip.show_options["knots"] = False

    item_to_show = [l_eye, r_eye, upper_lip, inner_mouth, lower_lip]
    if WITH_PUPILS:
        item_to_show += [l_pupil, r_pupil]

    plt = gus.show(
        item_to_show,
        lighting="off",
        background="white",
        close=False,
    )

    if WITH_DARK_BACKGROUND:
        background = []

        if DARK_WITH_WHITE_OVAL_BACKGROUND:
            height = 0.6
            width = 0

            x_shift = 0.3
            y_shift = -0.02
            cps = np.array([[0.0, 0.0], [0.0, height]])
            weights = np.ones([cps.shape[0]])
            knots = np.repeat([0.0, 1.0], 2)
            proto = splinepy.NURBS(
                degrees=[1],
                control_points=cps,
                weights=weights,
                knot_vectors=[knots],
            )
            left_half_sphere = proto.create.revolved(angle=180)
            right_half_sphere = proto.create.revolved(angle=-180)

            background = [
                left_half_sphere,
                right_half_sphere,
                splinepy.helpme.create.box(width + 2 * x_shift, 2 * height),
            ]
            for spline in background:
                spline.show_options["control_points"] = False
                spline.show_options["knots"] = False

            background[0].cps += [-x_shift, y_shift]
            background[1].cps += [width + x_shift, y_shift]
            background[2].cps += [-x_shift, -height + y_shift]
            for elem in background:
                elem.show_options["c"] = "white"
                if WITH_ALPHA_BACKGROUND:
                    elem.show_options["alpha"] = 0.25
        else:
            l_eye.show_options["c"] = "white"
            r_eye.show_options["c"] = "white"
            l_pupil.show_options["c"] = "black"
            r_pupil.show_options["c"] = "black"

        plt = gus.show(
            [*background, *item_to_show],
            lighting="off",
            background="black",
            close=False,
        )



from splinepy.helpme import (
    create,
    extract,
    fit,
    integrate,
    mapper,
    multi_index,
    reparametrize,
    visualize,
)

__all__ = [
    "create",
    "extract",
    "integrate",
    "mapper",
    "multi_index",
    "reparametrize",
    "visualize",
    "fit",
]



import argparse
import contextlib
from importlib import import_module

import splinepy
from splinepy._version import __version__


def convert(args):
    """Commandline interface functionality keyword convert.

    Convert splines from one format to another. See the io module for more information
    on what formats are supported for import and export.

    Args:
        args (namespace): Args namespace from argparse containing the options for
        convert.
    """
    fname = args.input_file
    from splinepy import io

    loaded = io.load(fname)
    if not isinstance(loaded, list):
        loaded = [loaded]
    if args.output_file:
        io.export(args.output_file, loaded)


def show(args):
    """Commandline interface functionality keyword show.

    Can load and show the file that is given in the args. Some options on what is shown
    can also be set.

    Args:
        args (namespace): Args namespace from argparse containing the options for show.
    """
    fname = args.input_file
    from splinepy import io

    loaded = io.load(fname)
    interactive = args.interactive or not args.output_file
    close = not bool(args.output_file)
    if interactive and not close:
        print(
            "Interactive mode is on and output file is set. "
            "Output file will be ignored. Instead, use S (Shift+S) in the "
            "interactive window to save a screenshot "
            "(saved as screenshot.png.)"
        )
    if not isinstance(loaded, list):
        loaded = [loaded]
    for loaded_object in loaded:
        loaded_object.show_options["knots"] = args.no_knot_vectors
        loaded_object.show_options["control_points"] = args.no_control_points
        loaded_object.show_options["control_point_ids"] = args.no_cp_ids
        loaded_object.show_options["control_mesh"] = args.no_control_mesh
        if args.resolution:
            loaded_object.show_options["resolutions"] = int(args.resolution)
    show_vedo = splinepy.show(loaded, interactive=interactive, close=close)
    if args.output_file:
        show_vedo.screenshot(args.output_file)


def entry():
    """Entry point for splinepy commandline interface.

    Currently only supports version and show commands.

    Try out `splinpy -h` for more information.
    """
    print("")
    print(
        "                    %%\\ %%\\                                         "
    )
    print(
        "                    %% |\\__|                                        "
    )
    print(
        " %%%%%%%\\  %%%%%%\\  %% |%%\\ %%%%%%%\\   %%%%%%\\   %%%%%%\\  %%\\ "
        "  %%\\ "
    )
    print(
        "%%  _____|%%  __%%\\ %% |%% |%%  __%%\\ %%  __%%\\ %%  __%%\\ %% |  %% |"
    )
    print(
        "\\%%%%%%\\  %% /  %% |%% |%% |%% |  %% |%%%%%%%% |%% /  %% |%% |  %% |"
    )
    print(
        " \\____%%\\ %% |  %% |%% |%% |%% |  %% |%%   ____|%% |  %% |%% |  %% |"
    )
    print(
        "%%%%%%%  |%%%%%%%  |%% |%% |%% |  %% |\\%%%%%%%\\ %%%%%%%  |\\%%%%%%% |"
    )
    print(
        "\\_______/ %%  ____/ \\__|\\__|\\__|  \\__| \\_______|%%  ____/  \\____%% |"
    )
    print(
        "          %% |                                  %% |      %%\\   %% |"
    )
    print(
        "          %% |                                  %% |      \\%%%%%%  |"
    )
    print(
        "          \\__|                                  \\__|       \\______/"
    )
    print("")
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(help="sub-command help", dest="command")
    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        help="Show version of splinepy.",
    )

    parser_plot = subparsers.add_parser("show", help="Show the given spline.")
    parser_plot.add_argument(
        "-i", "--input-file", type=str, help="Input file name.", required=True
    )
    parser_plot.add_argument(
        "-o",
        "--output-file",
        type=str,
        help="Export graphic to file.",
        required=False,
    )
    parser_plot.add_argument(
        "-e",
        "--interactive",
        action="store_true",
        help="Show graphic in interactive window even if save to file is on.",
    )
    parser_plot.add_argument(
        "-c",
        "--no-control_points",
        action="store_false",
        help="Do not show control points.",
    )
    parser_plot.add_argument(
        "-k",
        "--no-knot-vectors",
        action="store_false",
        help="Do not show knot vectors.",
    )
    parser_plot.add_argument(
        "-n",
        "--no-cp-ids",
        action="store_false",
        help="Do not show control point ids.",
    )
    parser_plot.add_argument(
        "-m",
        "--no-control-mesh",
        action="store_false",
        help="Do not show control mesh.",
    )
    parser_plot.add_argument(
        "-r",
        "--resolution",
        action="store",
        help="Resolution if the spline. Number of interpolated points.",
    )
    parser_convert = subparsers.add_parser(
        "convert", help="Convert the given spline."
    )
    parser_convert.add_argument(
        "-i", "--input-file", type=str, help="Input file name.", required=True
    )
    parser_convert.add_argument(
        "-o",
        "--output-file",
        type=str,
        help="Export graphic to file.",
        required=True,
    )
    args = parser.parse_args()

    if args.version:
        dependent_versions = [
            "numpy",
            # "funi", # has currently no __version__ attribute
            "gustaf",
            "vedo",
            "scipy",
            "napf",
            "meshio",
            "k3d",
        ]
        print(f"splinepy version: {__version__}")
        for lib in dependent_versions:
            version = "Not Installed"
            with contextlib.suppress(ImportError):
                module = import_module(lib)
                version = module.__version__
            print(f"{lib} version: {version}")
        return

    if args.command == "show":
        if args.input_file is None:
            parser_plot.print_help()
            return
        show(args)
    elif args.command == "convert":
        if args.input_file is None or args.output_file is None:
            parser_convert.print_help()
            return
        convert(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    entry()



from splinepy import settings as _settings
from splinepy import splinepy_core as _splinepy_core
from splinepy.bezier import BezierBase as _BezierBase


class RationalBezier(_BezierBase):
    r"""
    RationalBezier (Spline).

    Passes all arguments to :code:`super.__init__()`, see :class:`.BezierBase`.

    Rational Beziers are an extension to classical Beziers, similar like NURBS
    are an extension to B-Splines. By introducing a scalar weighting function
    :math:`W:\mathbb{R}^{N_{param}}\to\mathbb{R}^{+}_{*}`, we can construct
    modified (rational) basis functions, that allow to define rational
    Beziers.
    For a one-dimensional parameter space (:math:`N_{param}=1`), this
    weighting function reads

    .. math::
            W(u) = \sum_{i=0}^{l} B^{i;p}(u) w^i

    with scalar weights :math:`w^i\in\mathbb{R}^{+}_{*}`.
    Consequently, for the two-dimensional case (:math:`N_{param}=2`) we have

    .. math::
            W(u,v) = \sum_{i=0}^{l} \sum_{j=0}^{m} B^{i;p}(u) B^{j;q}(v)
                w^{i,j}

    and for the three-dimensional case (:math:`N_{param}=3`)

    .. math::
            W(u,v,w) = \sum_{i=0}^{l} \sum_{j=0}^{m} \sum_{k=0}^{n}
                B^{i;p}(u) B^{j;q}(v) B^{k;r}(w) w^{i,j,k}

    We proceed with a description of how to construct rational Beziers:

    .. note::
        For simplicity, we restrict ourselves to the three most common types
        of splines in the following, namely curves, surfaces and volumes,
        although :code:`splinepy` also supports higher dimensions, see
        the documentation of :class:`.Spline` for more information.

    1. A rational Bezier of degree :math:`p` with control points
    :math:`P^i\in\mathbb{R}^{N_{phys}}`, weights
    :math:`w^i\in\mathbb{R}^{+}_{*}`, and a one-dimensional parameter space
    (i.e., :math:`N_{param}=1`) corresponds to a line embedded into the
    physical space

    .. math::
            C(u) = \sum_{i=0}^{l} R^{i;p}(u) P^i

    with the modified (rational) basis functions

    .. math::
            R^{i;p}(u) = \frac{B^{i;p}(u) w^i}{W(u)}

    2. A rational Bezier of degrees :math:`p,q` with control points
    :math:`P^{i,j}\in\mathbb{R}^{N_{phys}}`, weights
    :math:`w^{i,j}\in\mathbb{R}^{+}_{*}`, and a two-dimensional parameter
    space (i.e., :math:`N_{param}=2`) corresponds to a surface embedded into
    the physical space

    .. math::
            S(u,v) = \sum_{i=0}^{l} \sum_{j=0}^{m} R^{i,j;p,q}(u,v) P^{i,j}

    with the modified (rational) basis functions

    .. math::
            R^{i,j;p,q}(u,v) =
                \frac{\tilde{N}^{i,j;p,q}(u,v) w^{i,j}}{W(u,v)}

    3. A rational Bezier of degrees :math:`p,q,r` with control points
    :math:`P^{i,j,k}\in\mathbb{R}^{N_{phys}}`, weights
    :math:`w^{i,j,k}\in\mathbb{R}^{+}_{*}`, and a three-dimensional parameter
    space (i.e., :math:`N_{param}=3`) corresponds to a volume embedded into
    the physical space

    .. math::
            V(u,v,w) = \sum_{i=0}^{l} \sum_{j=0}^{m} \sum_{k=0}^{n}
                R^{i,j,k;p,q,r}(u,v,w) P^{i,j,k}

    with the modified (rational) basis functions

    .. math::
            R^{i,j,k;p,q,r}(u,v,w) =
                \frac{
                    \tilde{N}^{i,j,k;p,q,r}(u,v,w) w^{i,j,k}
                }{
                    W(u,v,w)
                }

    Higher-dimensional instances are constructed accordingly.

    **Usage**:

    .. code-block:: python

        # Rational bezier curve (quarter arc)
        rational_bezier = splinepy.RationalBezier(
            degrees=[2],
            control_points=[[0.0, 0.0], [1.0, 0.0], [1.0, 1.0]],
            weights=[1.0, 2.0**-0.5, 1.0],
        )

    Parameters
    -----------
    degrees: (para_dim,) array-like
    control_points: (m, dim) array-like
    weights: (m) array-like
    spline: Spline

    Returns
    --------
    None
    """

    __slots__ = ()

    def __init__(
        self, degrees=None, control_points=None, weights=None, spline=None
    ):
        super().__init__(
            spline=spline,
            degrees=degrees,
            control_points=control_points,
            weights=weights,
        )

    @property
    def rationalbezier(self):
        return self.copy(saved_data=False)

    @property
    def nurbs(self):
        return _settings.NAME_TO_TYPE["NURBS"](
            spline=_splinepy_core.same_spline_with_knot_vectors(self)
        )



import numpy as _np

from splinepy.bezier import Bezier as _Bezier
from splinepy.microstructure.tiles.tile_base import TileBase as _TileBase


class Cross2D(_TileBase):
    """Simple crosstile with linear-quadratic branches and a bilinear
    center spline.

    .. raw:: html

        <p><a href="../_static/Cross2D.html">Fullscreen</a>.</p>
        <embed type="text/html" width="100%" height="400" src="../_static/Cross2D.html" />

    """  # noqa: E501

    _dim = 2
    _para_dim = 2
    _evaluation_points = _np.array(
        [
            [0.0, 0.5],
            [1.0, 0.5],
            [0.5, 0.0],
            [0.5, 1.0],
        ]
    )
    _n_info_per_eval_point = 1

    def _closing_tile(
        self,
        parameters=None,
        parameter_sensitivities=None,
        closure=None,
        boundary_width=0.1,
        filling_height=0.5,
        **kwargs,  # noqa ARG002
    ):
        """Create a closing tile to match with closed surface.

        Parameters
        ----------
        parameters : np.ndarray(4, 1)
          Four evaluation poinst with one parameter is used. This parameter
          describes the radius of the cylinder at the evaluation point.
          The parameters must be a two-dimensional np.array, where the
          value must be between 0.01 and 0.49
        parameter_sensitivities: np.ndarray
          Describes the parameter sensitivities with respect to some design
          variable. In case the design variables directly apply to the
          parameter itself, they evaluate as delta_ij
        boundary_width : float
          with of the boundary surrounding branch
        filling_height : float
          portion of the height that is filled in parametric domain
        closure : str
          parametric dimension that needs to be closed, given in the form
          "x_min", "x_max", etc.

        Returns
        -------
        list_of_splines : list
        derivative_list : list / None
        """
        # Check parameters
        if closure is None:
            raise ValueError("No closing direction given")

        if parameters is None:
            self._logd("Tile request is not parametrized, setting default 0.2")
            parameters = _np.array(
                _np.ones(
                    (len(self._evaluation_points), self._n_info_per_eval_point)
                )
                * 0.2
            )

        self.check_params(parameters)

        if not (_np.all(parameters > 0) and _np.all(parameters < 0.5)):
            raise ValueError("Thickness out of range (0, .5)")

        if not (0.0 < float(boundary_width) < 0.5):
            raise ValueError("Boundary Width is out of range")

        if not (0.0 < float(filling_height) < 1.0):
            raise ValueError("Filling must  be in (0,1)")

        # Check if user requests derivative splines
        if parameter_sensitivities is not None:
            # Check format
            self.check_param_derivatives(parameter_sensitivities)

            n_derivatives = parameter_sensitivities.shape[2]
            derivatives = []
        else:
            n_derivatives = 0
            derivatives = None

        splines = []
        for i_derivative in range(n_derivatives + 1):
            # Constant auxiliary values
            if i_derivative == 0:
                fill_height = filling_height
                bound_width = boundary_width
                inv_bound_width = 1.0 - bound_width
                inv_fill_height = 1.0 - fill_height
                ctps_mid_height_top = (1 + fill_height) * 0.5
                ctps_mid_height_bottom = 1.0 - ctps_mid_height_top
                v_one_half = 0.5
                v_one = 1.0
                v_zero = 0.0
                parameters = parameters[:, 0]
            else:
                # Set constant values to zero for derivatives
                fill_height = 0.0
                bound_width = 0.0
                inv_bound_width = 0.0
                inv_fill_height = 0.0
                ctps_mid_height_top = 0.0
                ctps_mid_height_bottom = 0.0
                v_one_half = 0.0
                v_one = 0.0
                v_zero = 0.0
                parameters = parameter_sensitivities[:, 0, i_derivative - 1]

            spline_list = []
            if closure == "x_min":
                # Minimum x position
                branch_thickness = parameters[1]

                block0_ctps = _np.array(
                    [
                        [v_zero, v_zero],
                        [fill_height, v_zero],
                        [v_zero, bound_width],
                        [fill_height, bound_width],
                    ]
                )

                block1_ctps = _np.array(
                    [
                        [v_zero, bound_width],
                        [fill_height, bound_width],
                        [v_zero, inv_bound_width],
                        [fill_height, inv_bound_width],
                    ]
                )

                block2_ctps = _np.array(
                    [
                        [v_zero, inv_bound_width],
                        [fill_height, inv_bound_width],
                        [v_zero, v_one],
                        [fill_height, v_one],
                    ]
                )

                branch_ctps = _np.array(
                    [
                        [fill_height, bound_width],
                        [ctps_mid_height_top, v_one_half - branch_thickness],
                        [v_one, v_one_half - branch_thickness],
                        [fill_height, inv_bound_width],
                        [ctps_mid_height_top, v_one_half + branch_thickness],
                        [v_one, v_one_half + branch_thickness],
                    ]
                )

                spline_list.append(
                    _Bezier(degrees=[1, 1], control_points=block0_ctps)
                )

                spline_list.append(
                    _Bezier(degrees=[1, 1], control_points=block1_ctps)
                )

                spline_list.append(
                    _Bezier(degrees=[1, 1], control_points=block2_ctps)
                )

                spline_list.append(
                    _Bezier(degrees=[2, 1], control_points=branch_ctps)
                )
            elif closure == "x_max":
                # Maximum x position
                branch_thickness = parameters[0]

                block0_ctps = _np.array(
                    [
                        [inv_fill_height, v_zero],
                        [v_one, v_zero],
                        [inv_fill_height, bound_width],
                        [v_one, bound_width],
                    ]
                )

                block1_ctps = _np.array(
                    [
                        [inv_fill_height, bound_width],
                        [v_one, bound_width],
                        [inv_fill_height, inv_bound_width],
                        [v_one, inv_bound_width],
                    ]
                )

                block2_ctps = _np.array(
                    [
                        [inv_fill_height, inv_bound_width],
                        [v_one, inv_bound_width],
                        [inv_fill_height, v_one],
                        [v_one, v_one],
                    ]
                )

                branch_ctps = _np.array(
                    [
                        [0, v_one_half - branch_thickness],
                        [
                            ctps_mid_height_bottom,
                            v_one_half - branch_thickness,
                        ],
                        [inv_fill_height, bound_width],
                        [v_zero, v_one_half + branch_thickness],
                        [
                            ctps_mid_height_bottom,
                            v_one_half + branch_thickness,
                        ],
                        [inv_fill_height, inv_bound_width],
                    ]
                )

                spline_list.append(
                    _Bezier(degrees=[1, 1], control_points=block0_ctps)
                )

                spline_list.append(
                    _Bezier(degrees=[1, 1], control_points=block1_ctps)
                )

                spline_list.append(
                    _Bezier(degrees=[1, 1], control_points=block2_ctps)
                )

                spline_list.append(
                    _Bezier(degrees=[2, 1], control_points=branch_ctps)
                )
            elif closure == "y_min":
                # Minimum y position
                branch_thickness = parameters[3]

                block0_ctps = _np.array(
                    [
                        [v_zero, v_zero],
                        [bound_width, v_zero],
                        [v_zero, fill_height],
                        [bound_width, fill_height],
                    ]
                )

                block1_ctps = _np.array(
                    [
                        [bound_width, v_zero],
                        [inv_bound_width, v_zero],
                        [bound_width, fill_height],
                        [inv_bound_width, fill_height],
                    ]
                )

                block2_ctps = _np.array(
                    [
                        [inv_bound_width, v_zero],
                        [v_one, v_zero],
                        [inv_bound_width, fill_height],
                        [v_one, fill_height],
                    ]
                )

                branch_ctps = _np.array(
                    [
                        [bound_width, fill_height],
                        [inv_bound_width, fill_height],
                        [v_one_half - branch_thickness, ctps_mid_height_top],
                        [v_one_half + branch_thickness, ctps_mid_height_top],
                        [v_one_half - branch_thickness, v_one],
                        [v_one_half + branch_thickness, v_one],
                    ]
                )

                spline_list.append(
                    _Bezier(degrees=[1, 1], control_points=block0_ctps)
                )

                spline_list.append(
                    _Bezier(degrees=[1, 1], control_points=block1_ctps)
                )

                spline_list.append(
                    _Bezier(degrees=[1, 1], control_points=block2_ctps)
                )

                spline_list.append(
                    _Bezier(degrees=[1, 2], control_points=branch_ctps)
                )
            elif closure == "y_max":
                # Maximum y position
                branch_thickness = parameters[2]

                block0_ctps = _np.array(
                    [
                        [v_zero, inv_fill_height],
                        [bound_width, inv_fill_height],
                        [v_zero, v_one],
                        [bound_width, v_one],
                    ]
                )

                block1_ctps = _np.array(
                    [
                        [bound_width, inv_fill_height],
                        [inv_bound_width, inv_fill_height],
                        [bound_width, v_one],
                        [inv_bound_width, v_one],
                    ]
                )

                block2_ctps = _np.array(
                    [
                        [inv_bound_width, inv_fill_height],
                        [v_one, inv_fill_height],
                        [inv_bound_width, v_one],
                        [v_one, v_one],
                    ]
                )

                branch_ctps = _np.array(
                    [
                        [v_one_half - branch_thickness, v_zero],
                        [v_one_half + branch_thickness, v_zero],
                        [
                            v_one_half - branch_thickness,
                            ctps_mid_height_bottom,
                        ],
                        [
                            v_one_half + branch_thickness,
                            ctps_mid_height_bottom,
                        ],
                        [bound_width, inv_fill_height],
                        [inv_bound_width, inv_fill_height],
                    ]
                )

                spline_list.append(
                    _Bezier(degrees=[1, 1], control_points=block0_ctps)
                )

                spline_list.append(
                    _Bezier(degrees=[1, 1], control_points=block1_ctps)
                )

                spline_list.append(
                    _Bezier(degrees=[1, 1], control_points=block2_ctps)
                )

                spline_list.append(
                    _Bezier(degrees=[1, 2], control_points=branch_ctps)
                )
            else:
                raise NotImplementedError(
                    "Requested closing dimension is not supported"
                )

            if i_derivative == 0:
                splines = spline_list.copy()
            else:
                derivatives.append(spline_list)
        # Return results
        return (splines, derivatives)

    def create_tile(
        self,
        parameters=None,
        parameter_sensitivities=None,
        center_expansion=1.0,
        closure=None,
        **kwargs,  # noqa ARG002
    ):
        """Create a microtile based on the parameters that describe the branch
        thicknesses.

        Thickness parameters are used to describe the inner radius of the
        outward facing branches

        Parameters
        ----------
        parameters : np.array(4, 1)
          Four evaluation point with one parameter is used. This parameter
          describes the radius of the cylinder at the evaluation point.
          The parameters must be a two-dimensional np.array, where the
          value must be between 0.01 and 0.49
        parameter_sensitivities: list(np.ndarray)
          Describes the parameter sensitivities with respect to some design
          variable. In case the design variables directly apply to the
          parameter itself, they evaluate as delta_ij
        center_expansion : float
            thickness of center is expanded by a factor
        closure : str
          parametric dimension that needs to be closed, given in the form
          "x_min", "x_max", etc.

        Returns
        -------
        microtile_list : list(splines)
        derivative_list : list / None
        """

        if not isinstance(center_expansion, float):
            raise ValueError("Invalid Type")

        if not ((center_expansion > 0.5) and (center_expansion < 1.5)):
            raise ValueError("Center Expansion must be in (.5,1.5)")

        max_radius = min(0.5, (0.5 / center_expansion))

        # set to default if nothing is given
        if parameters is None:
            self._logd("Setting branch thickness to default 0.2")
            parameters = _np.array(
                _np.ones(
                    (len(self._evaluation_points), self._n_info_per_eval_point)
                )
                * 0.2
            )

        self.check_params(parameters)

        if not (_np.all(parameters > 0) and _np.all(parameters < max_radius)):
            raise ValueError(f"Thickness out of range (0, {max_radius})")

        self.check_param_derivatives(parameter_sensitivities)

        # Check if user requests derivative splines
        if parameter_sensitivities is not None:
            # Check format
            self.check_param_derivatives(parameter_sensitivities)

            n_derivatives = parameter_sensitivities.shape[2]
            derivatives = []
        else:
            n_derivatives = 0
            derivatives = None

        # Closure requested, pass to function
        if closure is not None:
            return self._closing_tile(
                parameters=parameters,
                parameter_sensitivities=parameter_sensitivities,
                closure=closure,
                **kwargs,
            )

        splines = []
        for i_derivative in range(n_derivatives + 1):
            # Constant auxiliary values
            if i_derivative == 0:
                [x_min_r, x_max_r, y_min_r, y_max_r] = parameters[:, 0]
                v_one_half = 0.5
                # center radius
                center_r = (
                    (x_min_r + x_max_r + y_min_r + y_max_r)
                    / 4.0
                    * center_expansion
                ).item()
                hd_center = 0.5 * (0.5 + center_r)
            else:
                [x_min_r, x_max_r, y_min_r, y_max_r] = (
                    parameter_sensitivities[:, :, i_derivative - 1]
                ).flatten()
                v_one_half = 0.0
                # center radius
                center_r = (
                    (x_min_r + x_max_r + y_min_r + y_max_r)
                    / 4.0
                    * center_expansion
                )
                hd_center = 0.5 * center_r

            # Init return value
            spline_list = []

            # Create the center-tile
            center_points = _np.array(
                [
                    [-center_r, -center_r],
                    [center_r, -center_r],
                    [-center_r, center_r],
                    [center_r, center_r],
                ]
            ) + _np.array([v_one_half, v_one_half])

            y_min_ctps = _np.array(
                [
                    [-y_min_r, -v_one_half],
                    [y_min_r, -v_one_half],
                    [-y_min_r, -hd_center],
                    [y_min_r, -hd_center],
                    [-center_r, -center_r],
                    [center_r, -center_r],
                ]
            ) + _np.array([v_one_half, v_one_half])

            y_max_ctps = _np.array(
                [
                    [-center_r, center_r],
                    [center_r, center_r],
                    [-y_max_r, hd_center],
                    [y_max_r, hd_center],
                    [-y_max_r, v_one_half],
                    [y_max_r, v_one_half],
                ]
            ) + _np.array([v_one_half, v_one_half])

            x_min_ctps = _np.array(
                [
                    [-v_one_half, -x_min_r],
                    [-hd_center, -x_min_r],
                    [-center_r, -center_r],
                    [-v_one_half, x_min_r],
                    [-hd_center, x_min_r],
                    [-center_r, center_r],
                ]
            ) + _np.array([v_one_half, v_one_half])

            x_max_ctps = _np.array(
                [
                    [center_r, -center_r],
                    [hd_center, -x_max_r],
                    [v_one_half, -x_max_r],
                    [center_r, center_r],
                    [hd_center, x_max_r],
                    [v_one_half, x_max_r],
                ]
            ) + _np.array([v_one_half, v_one_half])

            spline_list.append(
                _Bezier(degrees=[1, 1], control_points=center_points)
            )

            spline_list.append(
                _Bezier(degrees=[2, 1], control_points=x_min_ctps)
            )

            spline_list.append(
                _Bezier(degrees=[2, 1], control_points=x_max_ctps)
            )

            spline_list.append(
                _Bezier(degrees=[1, 2], control_points=y_min_ctps)
            )

            spline_list.append(
                _Bezier(degrees=[1, 2], control_points=y_max_ctps)
            )

            # Pass to output
            if i_derivative == 0:
                splines = spline_list.copy()
            else:
                derivatives.append(spline_list)
        # Return results
        return (splines, derivatives)



import platform
import tempfile

import numpy as np

import splinepy


def test_iges_round_trip_bsplines(to_tmpf, are_splines_equal):
    """
    Test iges export-import routine
    """
    # temporarily disable windows - debug test (help wanted!)
    if platform.system().startswith(
        "Win"
    ) and splinepy.splinepy_core.build_type().startswith("debug"):
        splinepy.utils.log.warning(
            "Skipping io.iges test.",
            "Try release mode, if you need to use io.iges",
        )
        return None

    bsp_el2 = splinepy.BSpline(
        degrees=[1, 1],
        control_points=[[0, 1], [1, 1], [0, 2], [1, 2]],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
    )
    nur_el3 = splinepy.NURBS(
        degrees=[1, 1],
        control_points=[[1, 1], [2, 1], [1, 2], [2, 2]],
        weights=[1, 1, 1, 1],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
    )

    # Make it more tricky
    bsp_el2.elevate_degrees(0)
    bsp_el2.elevate_degrees(1)
    nur_el3.elevate_degrees(0)
    nur_el3.elevate_degrees(1)
    bsp_el2.insert_knots(1, [0.5])
    nur_el3.insert_knots(1, [0.5])

    # Test output against input
    list_of_splines = [bsp_el2, nur_el3]
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.iges.export(tmpf, list_of_splines)
        list_of_splines_loaded = splinepy.io.iges.load(tmpf)

        # to compare we need 3d "bumped" splines of originals
        splines_in_3d = []
        for s in list_of_splines:
            tmp_s = s.copy()
            tmp_s.cps = np.hstack([tmp_s.cps, np.zeros((len(tmp_s.cps), 1))])
            splines_in_3d.append(tmp_s)

        assert all(
            are_splines_equal(a, b)
            for a, b in zip(splines_in_3d, list_of_splines_loaded)
        )



from itertools import product

import numpy as np

import splinepy


def test_cartesian_product(np_rng):
    """Test cartesian product test"""

    # define arrays for cartesian product
    # make different len, but not too long so that test
    # doesn't take too long.
    max_dim = 10
    pool = np.array([0.0, 1.1, 4.4, 8.8, 9.6])
    queries = [pool[: np_rng.integers(2, 5)] for _ in range(max_dim)]

    # in case they are all the same, let's change two entries
    queries[-1] = pool
    queries[-2] = pool[:3]

    def ref(arrays, reverse):
        """create reference answers using itertools.product"""
        if reverse:
            prod = np.array(list(product(*arrays[::-1])))
            return prod[:, ::-1]
        else:
            prod = np.array(list(product(*arrays)))
            return prod

    # test cartesian product
    for i in range(1, max_dim + 1):
        in_arr = queries[:i]
        assert np.allclose(
            ref(in_arr, False),
            splinepy.utils.data.cartesian_product(in_arr, False),
        ), f"{i}-dim failed with NO reverse"

    # test reversed dim - what's often used in lib
    for i in range(1, max_dim + 1):
        in_arr = queries[:i]
        assert np.allclose(
            ref(in_arr, True),
            splinepy.utils.data.cartesian_product(in_arr, True),
        ), f"{i}-dim failed with reverse"



import gustaf as gus
import numpy as np

import splinepy


def bspline2p3d():
    """
    Creates new bspline of of parametric dimension 2, physical dimension 3
    """
    ds2 = [2, 2]

    # define knot vectors
    kvs2 = [
        [0.0, 0.0, 0.0, 0.5, 1.0, 1.0, 1.0],
        [0.0, 0.0, 0.0, 1.0, 1.0, 1.0],
    ]

    # define control points
    cps2 = np.array(
        [
            [0, 0, 0],
            [0, 1, 0],
            [1, 1.5, 0],
            [3, 1.5, 0],
            [-1, 0, 0],
            [-1, 2, 0],
            [1, 4, 0],
            [3, 4, 0],
            [-2, 0, 0],
            [-2, 2, 0],
            [1, 5, 0],
            [3, 5, 2],
        ]
    )

    # init bspline
    return splinepy.BSpline(
        degrees=ds2,
        knot_vectors=kvs2,
        control_points=cps2,
    )


if __name__ == "__main__":
    np_rng = np.random.default_rng()
    # turn on debug logs
    # gus.utils.log.configure(debug=True)

    # define spline_data
    # 1. see coordinates's norm
    b = bspline2p3d()
    b.spline_data["me"] = b
    b.show_options["data"] = "me"
    gus.show(["1. Show norm of coordinates.", b])

    # 1.1 default scalarbar
    b = bspline2p3d()
    b.spline_data["me"] = b
    b.show_options["data"] = "me"
    b.show_options["scalarbar"] = True
    gus.show(["1.1. Show 1. plus scalarbar", b])

    # 2. see coordinate's norm and as arrow
    b = bspline2p3d()
    b.spline_data["me"] = b
    b.show_options["data"] = "me"
    b.show_options["arrow_data"] = "me"
    gus.show(
        ["2. Show coordinate norm as scalar field and coordinate as arrows", b]
    )

    # 3. see coordinates norm and as arrows only on specified places
    b = bspline2p3d()
    b.spline_data["me"] = b
    b.show_options["data"] = "me"
    b.show_options["arrow_data"] = "me"
    b.show_options["arrow_data_on"] = np_rng.random((100, 2))  # para_coords
    gus.show(
        ["3. Show coordinates norm and as arrows on 100 random points.", b]
    )

    # 4. see 3. with parametric_view
    b = bspline2p3d()
    b.spline_data["me"] = b
    b.show_options["data"] = "me"
    b.show_options["arrow_data"] = "me"
    b.show_options["arrow_data_on"] = np_rng.random((100, 2))  # para_coords
    b.show_options["scalarbar"] = True
    gus.show(
        ["4. Show 3. and 3. in parametric space view with scalarbar.", b],
        b.create.parametric_view(),
    )
    # b.show(parametric_space=True)  # does the same

    # 5. plot function defines how to evaluate at given parametric location
    def plot_func(data, on):
        """
        callback to evaluate derivatives
        """
        return data.derivative(on, [0, 1])

    # use adaptor to defie a data
    plot_func_data = splinepy.SplineDataAdaptor(b, function=plot_func)
    # the rest is the same
    b = bspline2p3d()
    b.spline_data["der01"] = plot_func_data
    b.show_options["arrow_data"] = "der01"
    gus.show(
        [
            "5. Show partial derivative of seconda parametric dimension.\n"
            "This uses a callback function and SplineDataAdaptor.\n"
            "Also achievable with derivative spline.",
            b,
        ]
    )

    # remove on to sample same way as spline.
    # however, gold
    b = bspline2p3d()
    b.spline_data["der01"] = plot_func_data
    b.show_options["arrow_data"] = "der01"
    b.show_options["arrow_data_on"] = np_rng.random((100, 2))  # para_coords
    b.show_options["arrow_data_color"] = "gold"
    gus.show(
        [
            "5.1. Same data as 5. However, on 100 random places with gold "
            "arrows",
            b,
        ]
    )

    # 6. plot on predefined place - this will be only available as arrow data
    locations = np.repeat(np.linspace(0, 1, 15), 2).reshape(-1, 2)
    values = np.repeat(np.linspace(1, 2, 15), 3).reshape(-1, 3) + [0, 0, 2]
    fixed_data = splinepy.SplineDataAdaptor(values, locations=locations)
    b = bspline2p3d()
    b.spline_data["fixed"] = fixed_data
    b.show_options["arrow_data"] = "fixed"
    gus.show(
        [
            "6. Show arbitrary array data on predefined locations using "
            "SplineDataAdaptor.",
            b,
        ]
    )

    # fixed location data can't be shown in other requested locations.
    # following won't work
    # b.show_options["arrow_data_on"] = locations[:5]
    # b.show()

    # 7. plot any data with a function
    # some manually defined deformed spline
    deformed = bspline2p3d()  # minimal copy - properties and cached data only
    deformed.cps[11, -1] -= 4
    deformed.cps *= [5.5, 5.5, 5.5]
    deformed.cps += [-5, 0, 8]
    deformed.cps[0, [0, -1]] += 4
    deformed.show_options["c"] = "hotpink"

    # define callback
    def func(self_and_deformed, on):
        """
        callback to sample displacements.
        """
        # unpack data
        self, deformed = self_and_deformed
        return deformed.evaluate(on) - self.evaluate(on)

    # use adaptor - data is used as the first arg for callback
    deformed_data = splinepy.SplineDataAdaptor(
        data=(b, deformed),
        function=func,
    )
    b.spline_data["deformed"] = deformed_data
    b.show_options["arrow_data"] = "deformed"
    # arrows are always automatically scaled. for this one, let's not
    b.show_options["arrow_data_scale"] = 1
    b.show_options["arrow_data_on"] = locations
    # let's see in parametric space
    p_view = b.create.parametric_view()  # shallow copies data and options
    p_view.show_options.pop("arrow_data_scale")  # we want automatic scaling
    p_view.show_options["arrow_data_color"] = "gold"
    p_view.show_options["data"] = "deformed"
    # plot side by side
    gus.show(
        [
            "7. Original spline, deformed spline\n"
            "and pointwise correspondence on selected locations.",
            b,
            deformed,
        ],
        [
            "Parametric view of displacements.\n"
            "Arrows are directly transferred with automatic rescaling.\n"
            "(no inverse mapping)",
            p_view,
        ],
    )
    # say, deformed has changed again - plots should adapt automatically
    deformed.cps[:, [1, 2]] = deformed.cps[:, [2, 1]]
    gus.show(
        [
            "7.1. Same setup as 7. with inplace changes in deformed spline.",
            b,
            deformed,
        ],
        [
            "Parametric view of displacements.\n"
            "Arrows are directly transferred with automatic rescaling.\n"
            "(no inverse mapping)",
            p_view,
        ],
    )

    # 8. fixed location data that uses callback
    # predefind some locations
    bottom = splinepy.helpme.create.arc(
        radius=0.5, angle=-180
    )  # zero centered
    bottom.cps += [0.5, 0.55]
    circle1 = splinepy.helpme.create.circle(radius=0.1)
    circle2 = circle1.copy()
    circle1.cps += [0.25, 0.75]
    circle2.cps += [0.75, 0.75]
    n = 30
    locations = np.vstack(
        (bottom.sample(n), circle1.sample(n), circle2.sample(n))
    )

    # callback
    def heights(bot_c1_c2_n):
        bot, c1, c2, n_repeat = bot_c1_c2_n

        values = np.repeat(
            [
                [0, 0, bot],
                [0, 0, c1],
                [0, 0, c2],
            ],
            n_repeat,
            axis=0,
        )
        return values

    nice_data = splinepy.SplineDataAdaptor(
        data=(4, 2, 1, n),
        locations=locations,
        function=heights,
    )
    disc = splinepy.helpme.create.disk(2, angle=123)
    disc.normalize_knot_vectors()
    disc.spline_data["nice"] = nice_data
    disc.show_options["arrow_data"] = "nice"
    disc.show_options["arrow_data_color"] = "jet"
    gus.show(
        [
            "8. Showing arbitrary data with callback.\n"
            "Any array-like multi-dim data can be plotted as long as\n"
            "len(data) == len(locations) holds.",
            disc,
        ]
    )

    # 9. plot derivative
    bezier = splinepy.Bezier(
        degrees=[3], control_points=[[0, 0], [1, 0], [1, 2], [3, 2]]
    )
    bezier.spline_data["deriv"] = bezier.derivative_spline([1])
    bezier.show_options["arrow_data"] = "deriv"
    bezier.show_options["arrow_data_on"] = np.linspace(0, 1, 7).reshape(-1, 1)
    bezier.show_options["control_points"] = True
    bezier.show_options["arrow_data_scale"] = 0.3
    bezier.show_options["control_mesh"] = False
    bezier.show_options["control_point_ids"] = False
    splinepy.show(
        [
            "9. Show the derivative of the of the line as a vector plor\n"
            "Close form derivatives can be computed for Bezier types",
            bezier,
        ]
    )



"""
Parallel evaluation of spline, which requires to pickle spline.
"""

from functools import partial
from multiprocessing import Pool
from time import perf_counter as tic

import numpy as np

import splinepy

if __name__ == "__main__":
    np_rng = np.random.default_rng()
    # define a bspline
    ds = [2, 2]
    kvs = [
        [0, 0, 0, 0.5, 1, 1, 1],
        [0, 0, 0, 1, 1, 1],
    ]
    cps = [
        [0, 0, 0],
        [0, 1, 0],
        [1, 1.5, 0],
        [3, 1.5, 0],
        [-1, 0, 0],
        [-1, 2, 0],
        [1, 4, 0],
        [3, 4, 0],
        [-2, 0, 0],
        [-2, 2, 0],
        [1, 5, 0],
        [3, 5, 2],
    ]

    b = splinepy.BSpline(
        degrees=ds,
        knot_vectors=kvs,
        control_points=cps,
    )

    # elevate degree to make evaluation take a bit longer
    b.elevate_degrees(0)
    b.elevate_degrees(0)
    b.elevate_degrees(0)
    b.elevate_degrees(1)
    b.elevate_degrees(1)
    b.elevate_degrees(1)

    # let's try nurbs
    n = b.nurbs

    # workers
    workers = 4

    # form queries
    qs = [np_rng.random((5000, 2)) for _ in range(workers)]
    der = [2, 3]

    # full multiprocessing time
    now = tic()
    pool = Pool(workers)
    res_mp = list(pool.imap(partial(n.derivative, orders=der), qs))
    pool.close()
    pool.join()
    mptime0 = tic() - now
    print(f"multiprocess (with setup) evaluation took {mptime0} seconds.")

    # only evaluation multiprocessing time
    pool = Pool(workers)
    now = tic()
    res_mp = list(pool.imap(partial(n.derivative, orders=der), qs))
    mptime1 = tic() - now
    print(f"multiprocess (without setup) evaluation took {mptime1} seconds.")
    pool.close()
    pool.join()

    # stack batched queries
    q = np.vstack(qs)

    # multithread eval
    now = tic()
    res_t = n.derivative(q, der, nthreads=4)
    ttime = tic() - now
    print(f"multithread evaluation took {ttime} seconds.")

    # serial eval
    now = tic()
    res_s = n.derivative(q, der)
    stime = tic() - now
    print(f"serial evaluation took {stime} seconds.")

    print(f"serial / multiprocess (with setup) = {stime / mptime0}")
    print(f"serial / multiprocess (without setup) = {stime / mptime1}")
    print(f"serial / threading = {stime / ttime}")

    # cool, but are they correct?
    assert np.allclose(
        np.vstack(res_mp), res_s
    ), "Something went wrong during evaluation"
    assert np.allclose(res_t, res_s), "Something went wrong during evaluation"



import gustaf as gus

import splinepy as sp

if __name__ == "__main__":
    # 1.
    # create mesh 2d - extracted from bspline
    mesh_2d = sp.helpme.create.disk(
        outer_radius=3, inner_radius=2
    ).extract.faces([50, 30])
    spline_2d = sp.BSpline(
        [2, 2],
        [[0, 0, 0, 4, 4, 4], [2, 2, 2, 3, 3, 3]],
        [
            [0, 0],
            [0.5, -0.2],
            [1, 0],
            [0, 0.2],
            [0.5, 0.2],
            [1, 0.2],
            [0, 0.4],
            [0.5, 2],
            [1, 0.4],
        ],
    )

    ffd_2d = sp.FFD(mesh_2d, spline_2d)
    ffd_2d.show(title="2D FFD - BSpline")

    # 2.
    # Now 3D - again, create a mesh
    volume_3d = sp.helpme.create.torus(3, 2.5).extract.faces([30, 30, 30])

    # create spline
    spline_3d = sp.helpme.create.box(4.5, 2.5, 5.5)
    spline_3d.elevate_degrees([0, 1, 2])

    # manipulate cps
    # center cps
    cp_bounds = spline_3d.control_point_bounds
    spline_3d.control_points -= (cp_bounds[1] - cp_bounds[0]) / 2
    # use multi_index to get mid slice ids
    z_slice_ids = spline_3d.multi_index[:, :, 1]
    # rotate
    spline_3d.control_points[z_slice_ids] = gus.utils.arr.rotate(
        spline_3d.control_points[z_slice_ids], [0, 0, 79]
    )
    # translate
    spline_3d.control_points[z_slice_ids] += [0, 0, 0]

    ffd_3d = sp.FFD(volume_3d, spline_3d)
    ffd_3d.show(title="3D FFD - Bezier")

    # 3.
    # if you only provide mesh, we will generate a bounding spline for you.
    ffd_without_spline = sp.FFD(volume_3d)
    ffd_without_spline.show(title="Default bounding spline.")

    # then, you can just access spline and apply deformation
    ffd_without_spline.spline.control_points[0] -= 3
    ffd_without_spline.show(title="Manipulated default spline")

    # 4.
    # Only provide spline and then mesh after initialization
    # this will put spline
    # the result should be the same if you flip the setting
    ffd_with_out_mesh = sp.FFD()
    ffd_with_out_mesh.spline = spline_3d
    ffd_with_out_mesh.mesh = volume_3d
    ffd_with_out_mesh.show(title="Setting spline first, then mesh")



import contextlib
import os
import re

import numpy as np
import pytest

import splinepy


def error_log(*args):
    """used instead of print within the test"""
    splinepy.utils.log.error(*args)


@pytest.fixture
def np_rng():
    return np.random.default_rng()


# query points
@pytest.fixture
def queries_2D():
    return [
        [0.01, 0.01],
        [0.01, 0.5],
        [0.9, 0.1],
        [0.8, 0.7],
        [0.4, 0.99],
    ]


@pytest.fixture
def queries_3D():
    return [
        [0.1, 0.1, 0.1],
        [0.734, 0.525, 0.143],
        [0.9666, 0.991, 0.003],
        [0.5623, 0.0089, 0.99],
        [0.0431, 0.2, 0.523],
    ]


# initializing a spline should be a test itself, so provide `dict_spline`
# this is "iga-book"'s fig 2.15.
@pytest.fixture
def bspline_2p2d():
    return splinepy.BSpline(
        degrees=[2, 2],
        knot_vectors=[
            [0, 0, 0, 0.5, 1, 1, 1],
            [0, 0, 0, 1, 1, 1],
        ],
        control_points=[
            [0, 0],
            [0, 1],
            [1, 1.5],
            [3, 1.5],
            [-1, 0],
            [-1, 2],
            [1, 4],
            [3, 4],
            [-2, 0],
            [-2, 2],
            [1, 5],
            [3, 5],
        ],
    )


# half-half circle.
@pytest.fixture
def nurbs_2p2d():
    return splinepy.NURBS(
        degrees=[2, 1],
        knot_vectors=[
            [0, 0, 0, 1, 1, 1],
            [0, 0, 1, 1],
        ],
        control_points=[
            [-1.0, 0.0],
            [-1.0, 1.0],
            [0.0, 1.0],
            [-2.0, 0.0],
            [-2.0, 2.0],
            [0.0, 2.0],
        ],
        weights=[
            [1.0],
            [2**-0.5],
            [1.0],
            [1.0],
            [2**-0.5],
            [1.0],
        ],
    )


@pytest.fixture
def bezier_2p2d():
    return splinepy.Bezier(
        degrees=[2, 1],
        control_points=[
            [-1.0, 0.0],
            [-1.0, 1.0],
            [0.0, 1.0],
            [-2.0, 0.0],
            [-2.0, 2.0],
            [0.0, 2.0],
        ],
    )


@pytest.fixture
def rational_bezier_2p2d():
    return splinepy.RationalBezier(
        degrees=[2, 1],
        control_points=[
            [-1.0, 0.0],
            [-1.0, 1.0],
            [0.0, 1.0],
            [-2.0, 0.0],
            [-2.0, 2.0],
            [0.0, 2.0],
        ],
        weights=[
            [1.0],
            [2**-0.5],
            [1.0],
            [1.0],
            [2**-0.5],
            [1.0],
        ],
    )


@pytest.fixture
def bezier_3p3d():
    return splinepy.Bezier(
        degrees=[1, 1, 1],
        control_points=[
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, -1.0, 1.0],
            [1.0, 0.0, 1.0],
            [-1.0, 1.0, 2.0],
            [2.0, 2.0, 2.0],
        ],
    )


@pytest.fixture
def rational_bezier_3p3d():
    return splinepy.RationalBezier(
        degrees=[1, 1, 1],
        control_points=[
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, -1.0, 1.0],
            [1.0, 0.0, 1.0],
            [-1.0, 1.0, 2.0],
            [2.0, 2.0, 2.0],
        ],
        weights=[1.0] * 8,
    )


@pytest.fixture
def bspline_3p3d():
    return splinepy.BSpline(
        degrees=[1, 1, 1],
        control_points=[
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, -1.0, 1.0],
            [1.0, 0.0, 1.0],
            [-1.0, 1.0, 2.0],
            [2.0, 2.0, 2.0],
        ],
        knot_vectors=[
            [0.0, 0.0, 1.0, 1.0],
            [0.0, 0.0, 1.0, 1.0],
            [0.0, 0.0, 1.0, 1.0],
        ],
    )


@pytest.fixture
def nurbs_3p3d():
    return splinepy.NURBS(
        degrees=[1, 1, 1],
        control_points=[
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, -1.0, 1.0],
            [1.0, 0.0, 1.0],
            [-1.0, 1.0, 2.0],
            [2.0, 2.0, 2.0],
        ],
        weights=[1.0] * 8,
        knot_vectors=[
            [0.0, 0.0, 1.0, 1.0],
            [0.0, 0.0, 1.0, 1.0],
            [0.0, 0.0, 1.0, 1.0],
        ],
    )


@pytest.fixture
def raster():
    def _raster(bounds, resolutions):
        """prepares raster points using np.meshgrid"""
        l_bounds, u_bounds = bounds[0], bounds[1]
        pts = np.meshgrid(
            *[
                np.linspace(lo, up, re)
                for lo, up, re in zip(l_bounds, u_bounds, resolutions)
            ],
            indexing="ij",
        )
        # return pts
        return np.hstack([p.reshape(-1, 1) for p in pts[::-1]])

    return _raster


@pytest.fixture
def nd_box(raster):
    def _nd_box(dim):
        """creates simple box in n-d"""
        ds = [1 for _ in range(dim)]
        cps = raster(
            [[0 for _ in range(dim)], [1 for _ in range(dim)]],
            [2 for _ in range(dim)],
        )
        kvs = [[0, 0, 1, 1] for _ in range(dim)]
        ws = np.ones((len(cps), 1))
        return {
            "degrees": ds,
            "control_points": cps,
            "knot_vectors": kvs,
            "weights": ws,
        }

    return _nd_box


@pytest.fixture
def to_tmpf():
    def _to_tmpf(tmpd):
        """given tmpd, returns tmpf"""
        return os.path.join(tmpd, "nqv248p90")

    return _to_tmpf


@pytest.fixture
def are_splines_equal():
    def _are_splines_equal(a, b, print_=False):
        """returns True if Splines are equivalent"""
        if a.whatami != b.whatami:
            return False
        for req_prop in a.required_properties:
            if req_prop == "knot_vectors":
                for aa, bb in zip(a.knot_vectors, b.knot_vectors):
                    if not np.allclose(aa.numpy(), bb.numpy()):
                        if print_:
                            error_log("a.kvs", a.kvs)
                            error_log("b.kvs", b.kvs)
                        return False
            elif not np.allclose(getattr(a, req_prop), getattr(b, req_prop)):
                if print_:
                    error_log(f"a.{req_prop}", getattr(a, req_prop))
                    error_log(f"b.{req_prop}", getattr(b, req_prop))
                return False
        return True

    return _are_splines_equal


@pytest.fixture
def are_items_close():
    def _are_items_close(a, b):
        """returns True if items in a and b are close"""
        all_close = True

        for i, (aa, bb) in enumerate(zip(a, b)):
            if not all(np.isclose(aa, bb)):
                # print to inform
                error_log(f"elements in index-{i} are not close")
                error_log(f"  from first: {aa}")
                error_log(f"  from second: {bb}")

                all_close = False

        return all_close

    return _are_items_close


@pytest.fixture
def are_items_same():
    def _are_items_same(a, b):
        """returns True if items in a and b are same"""
        all_same = True

        for i, (aa, bb) in enumerate(zip(a, b)):
            if aa != bb:
                # print to inform
                error_log(f"element in index-{i} are not same")
                error_log(f"  from first: {aa}")
                error_log(f"  from second: {bb}")

                all_same = False

        return all_same

    return _are_items_same


@pytest.fixture
def are_stripped_lines_same():
    def _are_stripped_lines_same(a, b, ignore_order=False):
        """returns True if items in a and b same, preceding and tailing whitespaces
        are ignored and strings are joined"""
        all_same = True

        for i, (line_a, line_b) in enumerate(zip(a, b)):
            # check stripped string
            stripped_a, stripped_b = line_a.strip(), line_b.strip()

            # print general info
            if stripped_a != stripped_b:
                error_log(f"stripped line at index-{i} are not the same")
                error_log(f"  from first: {line_a}")
                error_log(f"  from second: {line_b}")

            # give one more chance if ignore_order
            if stripped_a != stripped_b and ignore_order:
                error_log("  checking again, while ignoring word order:")

                # This is meant for attributes
                delimiters = r" |\>|\<|\t|,"
                splitted_a = list(
                    filter(None, re.split(delimiters, stripped_a))
                )
                splitted_b = list(
                    filter(None, re.split(delimiters, stripped_b))
                )
                # first, len check
                len_a, len_b = len(splitted_a), len(splitted_b)
                if len(splitted_a) != len(splitted_b):
                    error_log(
                        f"    different word counts: a-{len_a}, b-{len_b}"
                    )
                    all_same = False
                else:
                    # word order
                    a_to_b = []
                    nums_b = None
                    for word_a in splitted_a:
                        try:
                            a_to_b.append(splitted_b.index(word_a))
                        except BaseException:
                            try:
                                num_a = float(word_a)
                            except ValueError:
                                pass
                            else:
                                if nums_b is None:
                                    nums_b = []
                                    for idx, num_b in enumerate(splitted_b):
                                        with contextlib.suppress(ValueError):
                                            nums_b.append((idx, float(num_b)))
                                for idx, num_b in nums_b:
                                    if np.isclose(num_a, num_b):
                                        a_to_b.append(idx)
                                        break
                                else:
                                    error_log(
                                        f"    second does not contain ({word_a})"
                                    )
                                    all_same = False

        return all_same

    return _are_stripped_lines_same



"""
irit io.
"""

import re as _re

import numpy as _np

from splinepy.io import ioutils as _ioutils

# Keywords that indicate new spline (irit is case insensitive, all lower case)
SPLINE_KEY_WORD_REGEX = r"\s+(bspline|bezier)\s+"

PARA_DIM_KEYS = {"1": "CURVE", "2": "SURFACE", "3": "TRIVAR"}


def load(fname, expand_tabs=True, strip_comments=False):
    """
    Read spline in `.itd` form.

    Parameters
    -----------
    fname: str
      Path to the irit file to read in.
    expand_tabs: bool
      Replace all tabs in the irit file. Defaults to True.
    strip_comments: bool
      (BETA) tries to identify comments and strips them from file

    Returns
    --------
    splines: list
      Spline Type defined in NAME_TO_TYPE
    """
    # Expand taps to spaces
    if expand_tabs:
        _ioutils.expand_tabs(fname)

    # Delete lines not containing brackets in the beginning
    def extract_relevant_text(lines):
        if strip_comments:
            return " ".join(
                [
                    line.lower().strip()
                    for line in lines
                    if _re.match(r".*[\[\]].*", line) is not None
                ]
            )
        else:
            return " ".join([line.lower().strip() for line in lines])

    with open(fname) as f:
        # Strip commentaries
        lines = f.readlines()
        relevant_text = extract_relevant_text(lines)

        # At this point we ignore all non-spline related data and do a keyword
        # based search for relevant information
        spline_strings = _re.split(SPLINE_KEY_WORD_REGEX, relevant_text)

        # Init list
        spline_list = []

        # Loop over text file data
        for type, data in zip(spline_strings[1::2], spline_strings[2::2]):
            # Dictionary of current spline
            spline = {}

            # Extract data from data
            extracted_data = _re.split(r"\]\s*\]", data)[0] + "]"

            # Split dimensions and degrees from data
            dim_and_deg = _re.split(r"(e\d|p\d)", extracted_data)

            # must contain ctps/degs rational/poly dimension control points
            if len(dim_and_deg) != 3:
                raise ValueError(
                    "Unsupported irit format, could not identify spline from "
                    "following string data set:\n"
                    + type.capitalize()
                    + " "
                    + extracted_data
                )

            # Dim is stored in second string written (E/P)+DIM
            dim = int(dim_and_deg[1][1:])
            is_rational = dim_and_deg[1][0] == "p"

            # Extract degrees (and control point dimensions)
            deg_info = [
                int(s) for s in _re.split(r"\s+", dim_and_deg[0]) if len(s) > 0
            ]

            # IRIT uses orders instead of degrees
            if type == "bezier":
                orders = _np.array(deg_info)
                n_ctps = _np.prod(orders)
            else:
                n_ctps = _np.prod(deg_info[: (len(deg_info) // 2)])
                orders = _np.array(deg_info[(len(deg_info) // 2) :])
            spline["degrees"] = orders - 1

            # Paradim is specified from degrees
            para_dim = orders.size

            # Extract control point and kv information
            ctps_and_kvs = _re.findall(r"\[(.*?)\]", dim_and_deg[2])
            ctps = " ".join(
                [s for s in ctps_and_kvs if _re.search(r"\s*kv.*", s) is None]
            )

            if is_rational:
                ctps = _np.fromstring(str(ctps), dtype=float, sep=" ").reshape(
                    -1, dim + 1
                )
                spline["weights"] = ctps[:, 0]
                spline["control_points"] = ctps[:, 1:] / ctps[:, 0:1]
            else:
                spline["control_points"] = _np.fromstring(
                    str(ctps), dtype=float, sep=" "
                ).reshape(-1, dim)

            if spline["control_points"].shape[0] != n_ctps:
                raise ValueError(
                    "Could not find sufficient number of control points in "
                    "spline:\n"
                    + type.capitalize()
                    + " "
                    + extracted_data
                    + "\nExpected "
                    + str(n_ctps)
                    + " got "
                    + str(spline["control_points"].shape[0])
                )

            # Extract knot-vectors
            if type == "bspline":
                # Extract knot-vectors
                knot_v_strings = [
                    s.split("kv")[1]
                    for s in ctps_and_kvs
                    if _re.search(r"\s*kv.*", s) is not None
                ]
                if len(knot_v_strings) != para_dim:
                    raise ValueError(
                        "Could not find enough knot vectors in bspline string:"
                        + type.capitalize()
                        + " "
                        + extracted_data
                        + "\nExpected "
                        + str(para_dim)
                        + " got "
                        + str(len(knot_v_strings))
                    )
                spline["knot_vectors"] = [
                    _np.fromstring(kv, dtype=float, sep=" ")
                    for kv in knot_v_strings
                ]

            # Append dictionary to list
            spline_list.append(spline)

        return _ioutils.dict_to_spline(spline_list)


def export(fname, splines):
    """
    Save splines as `.itd`.

    Parameters
    -----------
    fname: str
    splines: list

    Returns
    --------
    None
    """
    from splinepy.bezier import BezierBase as _BezierBase
    from splinepy.bspline import BSplineBase as _BSplineBase
    from splinepy.multipatch import Multipatch as _Multipatch
    from splinepy.spline import Spline as _Spline

    if isinstance(splines, _Multipatch):
        splines = splines.patches
    elif isinstance(splines, _Spline):
        splines = [splines]

    # Start Export
    with open(fname, "w") as f:
        # Header
        f.write("[OBJECT\n")

        # Elements
        for spline in splines:
            # First line contains information on control point mesh dimensions
            # and orders
            f.write("  [")
            # Paradim key
            f.write(PARA_DIM_KEYS.get(str(spline.para_dim), "MULTIVAR"))
            # Type identifier
            if isinstance(spline, _BezierBase):
                f.write(" BEZIER ")
            # Write orders or/and control point mesh resolutions
            if isinstance(spline, _BSplineBase):
                f.write(" BSPLINE ")
                f.write(
                    " ".join(
                        [
                            str(s)
                            for s in spline.control_mesh_resolutions.tolist()
                        ]
                    )
                    + " "
                )
            # Orders
            f.write(" ".join([str(d) for d in (spline.degrees + 1).tolist()]))

            # Write dimensionality
            if spline.is_rational:
                f.write(" P" + str(spline.dim) + "\n")
            else:
                f.write(" E" + str(spline.dim) + "\n")

            # Write knot vectors where required
            if isinstance(spline, _BSplineBase):
                for kv in spline.knot_vectors:
                    f.write(
                        "    [KV " + " ".join([str(k) for k in kv]) + "]\n"
                    )

            # Determine weighted control points
            if spline.is_rational:
                control_points = _np.hstack(
                    (
                        spline.weights.reshape(-1, 1),
                        spline.weights.reshape(-1, 1) * spline.control_points,
                    )
                )
            else:
                control_points = spline.control_points

            # Write control points
            f.write(
                "\n".join(
                    [
                        "    [" + " ".join([str(x) for x in ctp]) + "]"
                        for ctp in control_points.tolist()
                    ]
                )
            )

            # SPLINE FOOTER
            f.write("\n  ]\n")

        # FILE FOOTER
        f.write("]")
